from distutils.extension import Extension
import os

PATH_TO_PKG = os.path.relpath(os.path.dirname(__file__))
SOURCES = ("pair_counter_rp.pyx", )
THIS_PKG_NAME = '.'.join(__name__.split('.')[:-1])


def get_extensions():

    names = [THIS_PKG_NAME + "." + src.replace('.pyx', '') for src in SOURCES]
    sources = [os.path.join(PATH_TO_PKG, srcfn) for srcfn in SOURCES]
    include_dirs = ['numpy']
    libraries = []
    language = 'c++'
    extra_compile_args = ['-Ofast']

    extensions = []
    for name, source in zip(names, sources):
        extensions.append(Extension(name=name,
            sources=[source],
            include_dirs=include_dirs,
            libraries=libraries,
            language=language,
            extra_compile_args=extra_compile_args))

    return extensions


#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-07
# Last Modified: 2018-05-07
from __future__ import absolute_import, division, print_function
__author__     = ['Victor Calderon']
__copyright__  = ["Copyright 2018 Victor Calderon"]
__email__      = ['victor.calderon@vanderbilt.edu']
__maintainer__ = ['Victor Calderon']
__all__        = [  "spherematch"]

## Import modules
import math
import numpy as np
from   scipy.spatial import cKDTree as KDT
from   cosmo_utils.utils import file_utils as fd
from   cosmo_utils.custom_exceptions import LSSUtils_Error

## Functions

## Spherical to Cartesian - Normal version
def _spherical_to_cartesian(ra, dec):
    """
    Converts spherical coordinates (ra, dec) to cartesian coordinates (x,y,z).

    Parameters
    ----------
    ra, dec : array_like
        Right Ascension and Declination of the object. Units in `degrees`.

    Returns
    ----------
    x, y, z : array_like
        Cartesian coordinates of the object
    """
    # Converting to radians
    ra_arr  = np.radians(ra )
    dec_arr = np.radians(dec)
    # Cartesian coordinates
    x = np.cos(ra_arr) * np.cos(dec_arr)
    y = np.sin(ra_arr) * np.cos(dec_arr)
    z = np.sin(dec_arr)

    return x, y, z

## Spherical to Cartesian - Fast version
def _spherical_to_cartesian_fast(ra, dec, nthreads):
    """
    Converts spherical coordinates (ra, dec) to cartesian coordinates (x,y,z).

    Parameters
    ----------
    ra, dec : array_like
        Right Ascension and Declination of the object. Units in `degrees`

    nthreads : int
        Number of threads to use for calculation

    Returns
    ----------
    x, y, z : array_like
        Cartesian coordinates of the object

    Notes
    ----------
    This is a `fast` version of converting between spherical and
    cartesian coordinates.
    """
    # Importing 'numexpr'
    import numexpr as ne
    # Constants
    pi = math.pi
    # Setting number of Threads
    ne.set_num_threads(nthreads)
    # Evaluating arrays
    ra_arr  = ne.evaluate('ra*pi/180.0')
    dec_arr = ne.evaluate('dec*pi/180.0')
    hold1   = ne.evaluate('cos(dec_arr)')
    # Cartesian coordinates
    x = ne.evaluate('cos(ra_arr) * hold1')
    y = ne.evaluate('sin(ra_arr) * hold1')
    z = ne.evaluate('sin(dec_arr)')

    return x, y, z

## Great Circle - Normal version
def _great_circle_distance(ra1, dec1, ra2, dec2):
    """
    Computes the `great circle distance` between two points.

    Parameters
    ----------
    ra1, dec1 : array_like
        Right Ascension and Declination of the 1st point.
        Units in `degrees`.

    ra2, dec2 : array_like
        Right Ascension and Declination of the 2nd point.
        Units in `degrees`.

    Returns
    ----------
    great_circle_dist : float
        Great Circle distance between the 1st and 2nd location

    Notes
    ----------
    This function uses a vincenty distance.
    For more information see:
        https://en.wikipedia.org/wiki/Vincenty%27s_formulae

    It is a bit slower than others, but numerically stable.
    """
    # Importing modules
    from numpy import degrees, sin, cos, arctan2, hypot
    ##
    ## Terminology from the Vincenty Formula - `lambda` and `phi` and
    ## `standpoint` and `forepoint`
    lambs = np.radians(ra1 )
    phis  = np.radians(dec1)
    lambf = np.radians(ra2 )
    phif  = np.radians(dec2)
    # Calculations
    dlamb = lambf - lambs
    ## Constants for evaluation
    # Calculate these ones instead of few times!
    numera = cos(phif) * sin(dlamb)
    numerb = cos(phis) * sin(phif) - sin(phis) * cos(phif) * cos(dlamb)
    numer  = hypot(numera, numerb)
    denom  = sin(phis) * sin(phif) + cos(phis) * cos(phif) * cos(dlamb)
    # Great Circle Distance
    great_circle_dist = degrees(arctan2(numer, denom))

    return great_circle_dist

## Great Circle - Fast version
def _great_circle_distance_fast(ra1, dec1, ra2, dec2, nthreads):
    """
    Computes the `great circle distance` between two points.

    Parameters
    ----------
    ra1, dec1 : array_like
        Right Ascension and Declination of the 1st point.
        Units in `degrees`.

    ra2, dec2 : array_like
        Right Ascension and Declination of the 2nd point.
        Units in `degrees`.

    nthreads : int
        Number of threads to use for calculation

    Returns
    ----------
    great_circle_dist : float
        Great Circle distance between the 1st and 2nd location

    Notes
    ----------
    This function uses a vincenty distance.
    For more information see:
        https://en.wikipedia.org/wiki/Vincenty%27s_formulae

    It is a bit slower than others, but numerically stable.
    """
    import numexpr as ne
    ##
    ## Terminology from the Vincenty Formula - `lambda` and `phi` and
    ## `standpoint` and `forepoint`
    lambs = np.radians(ra1 )
    phis  = np.radians(dec1)
    lambf = np.radians(ra2 )
    phif  = np.radians(dec2)
    # Calculations
    dlamb = lambf - lambs
    ## Number of Threads
    ne.set_num_threads(nthreads)
    ## Constants for evaluation
    # Calculate these ones instead of few times!
    hold1  = ne.evaluate('sin(phif)' )
    hold2  = ne.evaluate('sin(phis)' )
    hold3  = ne.evaluate('cos(phif)' )
    hold4  = ne.evaluate('cos(dlamb)')
    hold5  = ne.evaluate('cos(phis)' )
    numera = ne.evaluate('hold3 * sin(dlamb)')
    numerb = ne.evaluate('hold5 * hold1 - hold2 * hold3 * hold4')
    numer  = ne.evaluate('sqrt(numera**2 + numerb**2)')
    denom  = ne.evaluate('hold2 * hold1 + hold5 * hold3 * hold4')
    pi     = math.pi
    # Great Circle Distance
    great_circle_dist = ne.evaluate('(arctan2(numer, denom))*180.0/pi')

    return great_circle_dist

## Sphere Match
def spherematch(ra1, dec1, ra2, dec2, tol=None, nnearest=1, nthreads=1):
    """
    Determines the matches between two catalogues of sources with
    <ra, dec> coordinates.

    Parameters
    ----------
    ra1, dec1 : array_like
        Right ascension and declination of the 1st catalogue.
        Units are in `degrees`.

    ra2, dec2 : array_like
        Right ascension and declination of the 2nd catalogue.
        Units are in `degrees`.

    tol : float or None, optional
        How close (in degrees) a match has to be to count as a match.
        If None, all nearest neighbors for the 1st catalogue will be returned.

    nnearest : int, optional
        The nth neighbor to find. E.g. 1 for the nearest nearby, 2 for the
        second nearest neighbor, etc. Partcularly useful if you want to get
        the nearest *non-self* neighbor of a catalogue.
        To do this use::

        ``spherematch(ra, dec, ra, dec, nnearest=2)``

        if `nnearest == 0`, all matches are returned.

    nthreads : int, optional
        Number of threads to use for calculation. This variable is set to
        1 by default. Must be larger than 1.

    Returns
    ----------
    idx1 : int `numpy.ndarray`
        Indices of the 1st catalogue of the matches. Will never be larger
        than `ra1`/`dec1`.

    idx2 : int `numpy.ndarray`
        Indices of the 2nd catalogue of the matches. Will never be larger
        than `ra1`/`dec1`.

    ds : float `numpy.ndarray`
        Distance (in degrees) between the matches.
    """
    file_msg = fd.Program_Msg(__file__)
    ## Checking input arguments
    valid_types = (list, np.ndarray)
    # `ra1`
    if not (isinstance(ra1, valid_types)):
        msg = '{0} `ra1` ({1}) is not a valid type!'.format(
            file_msg, type(ra1))
        raise LSSUtils_Error(msg)
    # `dec1`
    if not (isinstance(dec1, valid_types)):
        msg = '{0} `dec1` ({1}) is not a valid type!'.format(
            file_msg, type(dec1))
        raise LSSUtils_Error(msg)
    # `ra2`
    if not (isinstance(ra2, valid_types)):
        msg = '{0} `ra2` ({1}) is not a valid type!'.format(
            file_msg, type(ra2))
        raise LSSUtils_Error(msg)
    # `dec2`
    if not (isinstance(dec2, valid_types)):
        msg = '{0} `dec2` ({1}) is not a valid type!'.format(
            file_msg, type(dec2))
        raise LSSUtils_Error(msg)
    # `nnearest`
    if nnearest < 0:
        msg = '{0} `nnearest` ({1}) must be larger than `0`!'.format(
            file_msg,
            nnearest)
        raise LSSUtils_Error(msg)
    # `threads`
    if nthreads < 1:
        msg = '{0} `nthreads` ({1}) must be larger than `1`!'.format(
            file_msg,
            nthreads)
        raise LSSUtils_Error(msg)
    ##
    ## Converting arguments into arrays for ease of use
    ra1  = np.array(ra1 , copy=False)
    dec1 = np.array(dec1, copy=False)
    ra2  = np.array(ra2 , copy=False)
    dec2 = np.array(dec2, copy=False)
    ## Checking shape
    # 1st catalogue
    if ra1.shape != dec1.shape:
        msg = '{0} The shape of `ra1` ({1}) does not mathc that of `dec1` ({2}).'
        msg = msg.format(file_msg, ra1.shape, dec1.shape)
        raise LSSUtils_Error(msg)
    # 2nd catalogue
    if ra2.shape != dec2.shape:
        msg = '{0} The shape of `ra2` ({1}) does not mathc that of `dec2` ({2}).'
        msg = msg.format(file_msg, ra2.shape, dec2.shape)
        raise LSSUtils_Error(msg)
    ##
    ## Converting spherical coordinates into cartesian coordinates
    # 1st catalogue
    x1, y1, z1 = _spherical_to_cartesian_fast(  ra1.ravel(),
                                                dec1.ravel(),
                                                nthreads)
    coords1 = np.empty((x1.size,3))
    coords1[:, 0] = x1
    coords1[:, 1] = y1
    coords1[:, 2] = z1
    # 2nd catalogue
    x2, y2, z2 = _spherical_to_cartesian_fast(  ra2.ravel(),
                                                dec2.ravel(),
                                                nthreads)
    coords2 = np.empty((x2.size,3))
    coords2[:, 0] = x2
    coords2[:, 1] = y2
    coords2[:, 2] = z2
    ##
    ## Finding nearest neighbors
    kdt = KDT(coords2)
    # Finding neighbors
    if nnearest == 1:
        idx_s2 = kdt.query(coords1)[1]
    elif (nnearest == 0) and (tol is not None): # if you want ALL matches!
        p1_x, p1_y, p1_z = _spherical_to_cartesian_fast(90., 0  , nthreads)
        p2_x, p2_y, p2_z = _spherical_to_cartesian_fast(90., tol, nthreads)
        # Converting to floats
        p1_x   = float(p1_x)
        p1_y   = float(p1_y)
        p1_z   = float(p1_z)
        p2_x   = float(p2_x)
        p2_y   = float(p2_y)
        p2_z   = float(p2_z)
        r      = np.sqrt((p2_x - p1_x)**2 + (p2_y - p1_y)**2 + (p2_z - p1_z)**2)
        idx_s2 = kdt.query_ball_point(coords1, r)[0]
    elif nnearest > 1:
        idx_s2 = kdt.query(coords1, nnearest)[1][:, -1]
    else:
        msg = '{0} Invalid `nnearest` ({1})!'.format(file_msg, nnearest)
        raise LSSUtils_Error(msg)
    ##
    ## Calculating distance between matches
    ds = _great_circle_distance_fast(   ra1         ,
                                        dec1        ,
                                        ra2[idx_s2] ,
                                        dec2[idx_s2],
                                        nthreads    )
    ##
    ## If `tol` is None, then all objects will have a match.
    idx_s1 = np.arange(ra1.size)
    ##
    ## Remove matches that are `beyond` the tolerance separation
    if (tol is not None) and (nnearest != 0):
        mask   = ds < tol
        idx_s1 = idx_s1[mask]
        idx_s2 = idx_s2[mask]
        ds     = ds    [mask]

    return idx_s1, idx_s2, ds



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-07
# Last Modified: 2018-05-07
from __future__ import absolute_import, division, print_function
__author__     = ['Victor Calderon']
__copyright__  = ["Copyright 2018 Victor Calderon"]
__email__      = ['victor.calderon@vanderbilt.edu']
__maintainer__ = ['Victor Calderon']
__all__        = [  "apparent_to_absolute_magnitude",
                    "absolute_to_apparent_magnitude",
                    "get_sun_mag",
                    "absolute_magnitude_to_luminosity",
                    "luminosity_to_absolute_mag",
                    "absolute_magnitude_lim"]

## Importing modules
import numpy as np
from   cosmo_utils.utils import file_utils as fd
from   cosmo_utils.custom_exceptions import LSSUtils_Error

## Functions

## Apparent to Absolute magnitudes
def apparent_to_absolute_magnitude(app_mag, lum_dist, unit='mpc'):
    """
    Calculates the absolute magnitude based on luminosity and apparent
    magnitude.

    Parameters
    -----------
    app_mag : array_like
        Array of apparent magnitude(s)

    lum_dist : array_like
        Array of luminosity distnace to object. In units of `Mpc`.

    unit : {'pc', 'kpc', 'mpc'} str, optional
        Unit to use for `lum_dist`. This variable is set to `mpc` by
        default. When `pc`, the units are in parsecs, while `mpc` is for
        distances in mega-parsecs, etc.

    Returns
    -----------
    abs_mag : np.ndarray
        Array of absolute magnitudes. `abs_mag` is a float if
        `app_mag` is a float or int.
    """
    file_msg = fd.Program_Msg(__file__)
    ## Checking input parameters
    valid_types = (float, np.ndarray, list, int)
    # Type for `app_mag`
    if not (isinstance(app_mag, valid_types)):
        msg = '{0} `app_mag` ({1}) is not a valid type!'.format(file_msg,
            type(app_mag))
        raise LSSUtils_Error(msg)
    # Type for `unit`
    unit_valid_arr = ['mpc', 'kpc', 'pc']
    if not (unit in unit_valid_arr):
        msg = '{0} `unit` ({1}) is not a valid input!'.format(file_msg, unit)
        raise LSSUtils_Error(msg)
    ## Converting to array-type
    # `app_mag` object
    if (isinstance(app_mag, float) or isinstance(app_mag, int)):
        app_mag = float(app_mag)
    if (isinstance(app_mag, list) or isinstance(app_mag, np.ndarray)):
        app_mag = np.asarray(app_mag)
    # `lum_dist` object
    if (isinstance(lum_dist, float) or isinstance(lum_dist, int)):
        lum_dist = float(lum_dist)
    if (isinstance(lum_dist, list) or isinstance(lum_dist, np.ndarray)):
        lum_dist = np.asarray(lum_dist)
    # Units - Conveting to Mpc
    # This follows the formula:
    #   app_mag - abs_mag = 5 * (np.log10(lum_dist) + a - 1)
    #       Where a = 0 when [d] = parsecs
    #       Where a = 3 when [d] = kiloparsecs
    #       Where a = 6 when [d] = megaparsecs
    if unit == 'pc':
        a = 0
    elif unit == 'kpc':
        a = 3
    elif unit == 'mpc':
        a = 6
    ##
    ## Calcualtions
    abs_mag = app_mag - 5 * (np.log10(lum_dist) - 1 + a)

    return abs_mag

## Absolute to Apparent magnitudes
def absolute_to_apparent_magnitude(abs_mag, lum_dist, unit='mpc'):
    """
    Calculates the apparent magnitude using the luminosity and absolute
    magnitude.

    Parameters
    -----------
    abs_mag : float, int, or array_like
        Array of absolute magnitude(s)

    lum_dist : array_like
        Array of luminosity distnace to object. In units of `Mpc`.

    unit : {'pc', 'kpc', 'mpc'} str, optional
        Unit to use for `lum_dist`. This variable is set to `mpc` by
        default. When `pc`, the units are in parsecs, while `mpc` is for
        distances in mega-parsecs, etc.

    Returns
    -----------
    app_mag : array_like, or float
        Array of apparent magnitude(s). `app_mag` is a float if
        `abs_mag` is a float or int.
    """
    file_msg = fd.Program_Msg(__file__)
    ## Checking input parameters
    valid_types = (float, np.ndarray, list, int)
    # Type for `abs_mag`
    if not (isinstance(abs_mag, valid_types)):
        msg = '{0} `abs_mag` ({1}) is not a valid type!'.format(file_msg,
            type(abs_mag))
        raise LSSUtils_Error(msg)
    # Type for `unit`
    unit_valid_arr = ['mpc', 'pc']
    if not (unit in unit_valid_arr):
        msg = '{0} `unit` ({1}) is not a valid input!'.format(file_msg, unit)
        raise LSSUtils_Error(msg)
    ## Converting to array-type
    # `abs_mag` object
    if (isinstance(abs_mag, float) or isinstance(abs_mag, int)):
        abs_mag = float(abs_mag)
    if (isinstance(abs_mag, list) or isinstance(abs_mag, np.ndarray)):
        abs_mag = np.asarray(abs_mag)
    # `lum_dist` object
    if (isinstance(lum_dist, float) or isinstance(lum_dist, int)):
        lum_dist = float(lum_dist)
    if (isinstance(lum_dist, list) or isinstance(lum_dist, np.ndarray)):
        lum_dist = np.asarray(lum_dist)
    # Units - Conveting to Mpc
    # This follows the formula:
    #   app_mag - abs_mag = 5 * (np.log10(lum_dist) + a - 1)
    #       Where a = 0 when [d] = parsecs
    #       Where a = 3 when [d] = kiloparsecs
    #       Where a = 6 when [d] = megaparsecs
    if unit == 'pc':
        a = 0
    elif unit == 'kpc':
        a = 3
    elif unit == 'mpc':
        a = 6
    ##
    ## Calcualtions
    app_mag = abs_mag + 5. * (np.log10(lum_dist) - 1 + a)

    return app_mag

## Sun's magnitudes
def get_sun_mag(filter_opt, system='SDSS_Blanton_2003_z0.1'):
    """
    Get solar absolaute magnitude for a filter in a system.
    Taken from Duncan Campbell, and later modified.

    Parameters
    ----------
    filter_opt : {'U', 'B', 'V', 'R', 'I', 'J', 'H', 'K'} str
        Magnitude filter to use.

    system : {'Binney_and_Merrifield_1998', 'SDSS_Blanton_2003_z0.1'} str
        Kind of filter to use.

        Options:
            - 'Binney_and_Merrifield_1998' : See Binney and Merrifield, 1998
            - 'SDSS_Blanton_2003_z0.1' : See Blanton et al. (2003) Eqn. 14.

    Returns
    ----------
    abs_mag_sun : float
        Solar absolute magnitude in `filter_opt` using `system` parameters.

    Raises
    ----------
    LSSUtils_Error : Exception
        Program exception if input parameters are accepted

    Examples
    ----------
    >>> get_sun_mag('R', 'Binney_and_Merrifield_1998')
    4.42

    >>> get_sun_mag('V', 'Binney_and_Merrifield_1998')
    4.83

    >>> get_sun_mag('g', 'SDSS_Blanton_2003_z0.1')
    5.45
    """
    file_msg = fd.Program_Msg(__file__)
    ## Checking input parameters
    filter_arr = ['U', 'B', 'V', 'R', 'I', 'J', 'H', 'K', 'u','g','r','i','z']
    system_arr = ['Binney_and_Merrifield_1998', 'SDSS_Blanton_2003_z0.1']
    # Checks
    # Input filter
    if not (filter_opt in filter_arr):
        msg = '{0} `filter_opt` ({1}) is not a valid option!'.format(
            file_msg, filter_opt)
        raise LSSUtils_Error
    # Input system
    if not (system in system_arr):
        msg = '{0} `system` ({1}) is not a valid option!'.format(
            file_msg, system)
        raise LSSUtils_Error
    ##
    ## Input parameters
    abs_mag_sun_dict = {'Binney_and_Merrifield_1998' : {'U': 5.61,
                                                        'B': 5.48,
                                                        'V': 4.83,
                                                        'R': 4.42,
                                                        'I': 4.08,
                                                        'J': 3.64,
                                                        'H': 3.32,
                                                        'K': 3.28},
                        'SDSS_Blanton_2003_z0.1'    : { 'u': 6.80,
                                                        'g': 5.45,
                                                        'r': 4.76,
                                                        'i': 4.58,
                                                        'z': 4.51}}
    ## Checking if key exists in dictionary
    ## and assigning magnitude
    if (filter_opt in abs_mag_sun_dict[system].keys()):
        abs_mag_sun = abs_mag_sun_dict[system][filter_opt]
    else:
        msg = '{0} `filter_opt` ({1}) is not a proper key of `system` ({2})'
        msg = msg.format(file_msg, filter_opt, system)
        raise LSSUtils_Error(msg)

    return abs_mag_sun

## Absolute magnitude to luminosity
def absolute_magnitude_to_luminosity(abs_mag, filter_opt,
    system='SDSS_Blanton_2003_z0.1'):
    """
    Calculates the luminosity of the object through `filter_opt` filter.

    Parameters
    -----------
    abs_mag : float, int, or array_like
        Absolute magnitude of one or multiple objects.

    filter_opt : {'U', 'B', 'V', 'R', 'I', 'J', 'H', 'K'} str
        Magnitude filter to use.

    system : {'Binney_and_Merrifield_1998', 'SDSS_Blanton_2003_z0.1'} str
        Kind of filter to use.

        Options:
            - 'Binney_and_Merrifield_1998' : See Binney and Merrifield, 1998
            - 'SDSS_Blanton_2003_z0.1' : See Blanton et al. (2003) Eqn. 14.

    Returns
    -----------
    log_L : float or array_like
        Logarithmic value of the luminosity in the `filter_opt` band.

    Raises
    ----------
    LSSUtils_Error : Exception
        Program exception if input parameters are accepted
    """
    file_msg = fd.Program_Msg(__file__)
    ## Checking input parameters
    valid_types = (float, int, list, np.ndarray)
    if not (isinstance(abs_mag, valid_types)):
        msg = '{0} `abs_mag` ({1}) is not a valid type!'.format(file_msg,
            abs_mag)
        raise LSSUtils_Error(msg)
    ## Obtaining Sun's absolute magnitude
    abs_mag_sun = get_sun_mag(filter_opt, system=system)
    ## Luminosity calculations
    log_L = (abs_mag_sun - abs_mag) * 0.4

    return log_L

## Luminosity to Absolute magnitude
def luminosity_to_absolute_mag(lum, filter_opt,
    system='SDSS_Blanton_2003_z0.1'):
    """
    Calculates the absolute magnitude of object through the `filter_opt`
    filter.

    Parameters
    -----------
    lum : float, int, array_like
        Luminosity of 1 or more objects. In units of `solar luminosities`.

    filter_opt : {'U', 'B', 'V', 'R', 'I', 'J', 'H', 'K'} str
        Magnitude filter to use.

    system : {'Binney_and_Merrifield_1998', 'SDSS_Blanton_2003_z0.1'} str
        Kind of filter to use.

        Options:
            - 'Binney_and_Merrifield_1998' : See Binney and Merrifield, 1998
            - 'SDSS_Blanton_2003_z0.1' : See Blanton et al. (2003) Eqn. 14.

    Returns
    -----------
    abs_mag : float, int, or array_like
        Absolute magnitude of one or multiple objects. Same type as `lum`

    Raises
    ----------
    LSSUtils_Error : Exception
        Program exception if input parameters are accepted
    """
    file_msg = fd.Program_Msg(__file__)
    ## Checking input parameters
    valid_types = (float, int, list, np.ndarray)
    if not (isinstance(lum, valid_types)):
        msg = '{0} `lum` ({1}) is not a valid type!'.format(file_msg,
            lum)
        raise LSSUtils_Error(msg)
    ## Obtaining Sun's absolute magnitude
    abs_mag_sun = get_sun_mag(filter_opt, system=system)
    ## Absolute magnitude calculation
    # In units of solar luminosities
    lum_sun = 1.0
    # Absolute magnitude of objects
    abs_mag = abs_mag_sun - 2.5 * np.log10(lum / lum_sun)

    return abs_mag

## Absolute magnitude limits
def absolute_magnitude_lim(z, mag_lim, cosmo=None, H0=100., verbose=True):
    """
    Calculates the absolute magnitude limit as function of redshift `z` for
    a flux-limited survey.

    Parameters
    -----------
    z : float, int, or array_like
        Maximum redshift for a given flux-limited survey.

    mag_lim : float
        Apparent magnitude limit of the flux-limited survey.

    cosmo : `astropy.cosmology` object
        Cosmology object from Astropy.

    H0 : float, optional
        Hubble parameters value used to estimate distances.
        This variable is set to 100 km/s/Mpc by default.

    verbose : `bool`, optional
        If True, a message will appear when the default cosmology is used.

    Returns
    -----------
    abs_mag : float, int, or array_like
        Absolute magnitude limit in units of `abs_mag` + 5*log10(h),
        where `h` is the little Hubble parameter.

    Raises
    ----------
    LSSUtils_Error : Exception
        Program exception if input parameters are accepted
    """
    file_msg = fd.Program_Msg(__file__)
    ## Checking input parameters
    # Redshift
    z_valid_types = (float, int, list, np.ndarray)
    if not (isinstance(z, z_valid_types)):
        msg = '{0} `z` ({1}) is not a valid type!'.format(file_msg, type(z))
        raise LSSUtils_Error(msg)
    # Magnitude limit
    mag_lim_valid_types = (float, int)
    if not (isinstance(mag_lim, mag_lim_valid_types)):
        msg = '{0} `mag_lim` ({1}) is not a valid type!'.format(file_msg,
            type(mag_lim))
        raise LSSUtils_Error(msg)
    # Hubble parameter value
    H0_valid_types = (float, int)
    if not (isinstance(H0, H0_valid_types)):
        msg = '{0} `H0` ({1}) is not a valid type!'.format(file_msg,
            type(H0))
        raise LSSUtils_Error(msg)
    ##
    ## Calculations
    if not cosmo:
        from astropy.cosmology import FlatLambdaCDM
        cosmo = FlatLambdaCDM(H0=H0, Om0=0.316)
        if verbose:
            print(">> Warning: No cosmology was specified. Using default:",
                cosmo)
    ## Luminosity distance
    lum_dist = cosmo.luminosity_distance(z).value
    ## Absolute magnitude
    abs_mag = apparent_to_absolute_magnitude(mag_lim, lum_dist)

    return abs_mag



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-09
# Last Modified: 2018-05-09
from __future__ import print_function, division, absolute_import
__author__     =['Victor Calderon']
__copyright__  =["Copyright 2018 Victor Calderon"]
__email__      =['victor.calderon@vanderbilt.edu']
__maintainer__ =['Victor Calderon']
__all__        =[   "pairwise_distance_rp"]

"""
Computes the number of pairs in perpendicular projected distance `rp` bins
for a set of galaxies with <ra dec cz> positions.
"""

## Importing modules
cimport cython
import numpy as np
cimport numpy as cnp
from libc.math cimport sqrt, log10, fabs
from cosmo_utils.custom_exceptions import LSSUtils_Error

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)

# Functions

# Pair Counter - `rp`
def pairwise_distance_rp(coord_1, coord_2, rpmin=0.01, rpmax=10,
    nrpbins=10, pimax=20., double_count=False):
    """
    Cython engine for returning pairs of points separated in projected
    radial bins with an observer at (0,0,0).

    Parameters
    -----------
    coord_1 : array_like, shape (N,3)
        Cartesian corodinates <x, y, z> of `sample 1`.

    coord_2 : array_like, shape (N,3)
        Cartesian corodinates <x, y, z> of `sample 2`.

    rpmin : `float`, optional
        Minimum `rp` (perpendicular distance) to search for and return pairs.
        This variable is set to `0.01` by default.

    rpmax : `float`, optional
        Maximum `rp` (perpendicular distance) to search for and return pairs.
        This variable is set to `10` by default.

    nrpbins : `int`, optional
        Total number of `rp` bins. This variable is set to `10` by default.

    pimax : `float`, optional
        Maximum parallel separation distance to search for and return pairs.
        This variable is set to `20` by default.

    double_count : `bool`, optional
        If True, the function double counts for each galaxy pairs.
        Useful for the `2-halo Quenched Fractions` calculations.

    Returns
    -----------
    rp_ith_arr : `numpy.ndarray`, shape (M, 3)
        Three-dimensional array of M-elements containing:
            - `rp` : bin number to which galaxy pair belongs.
            - `i_ind` : indices of 0-indexed indices in `sample 1`.
            - `j_ind` : indices of 0-indexed indices in `sample 2`.
    """
    # -- Output Lists --
    rp_arr  = []
    ith_arr = []
    jth_arr = []
    # Number of Elements
    cdef int Ni, Nj, i, j
    cdef cnp.float64_t rpbin
    Ni = len(coord_1)
    Nj = len(coord_2)
    # -- Constants --
    cdef cnp.float64_t PI180 = 3.141592653589793/180.
    # -- Count Pairs variables --
    cdef cnp.float64_t sx, sy, sz, lx, ly, lz, l2, ll, spar, s2, sperp
    # -- `rp` constants
    cdef int nrpbins_p          = nrpbins
    cdef cnp.float64_t rp_min_p = rpmin
    cdef cnp.float64_t rp_max_p = rpmax
    cdef cnp.float64_t pi_max_p = pimax
    cdef cnp.float64_t logrpmin = log10(rpmin)
    cdef cnp.float64_t logrpmax = log10(rpmax)
    cdef cnp.float64_t dlogrp   = (logrpmax-logrpmin)/nrpbins
    # -- Cartesian Coordinates --
    # Sample 1
    cdef cnp.float64_t[:] x1 = coord_1.astype(float).T[0]
    cdef cnp.float64_t[:] y1 = coord_1.astype(float).T[1]
    cdef cnp.float64_t[:] z1 = coord_1.astype(float).T[2]
    # Sample 2
    cdef cnp.float64_t[:] x2 = coord_2.astype(float).T[0]
    cdef cnp.float64_t[:] y2 = coord_2.astype(float).T[1]
    cdef cnp.float64_t[:] z2 = coord_2.astype(float).T[2]
    # Determining if to double count or not
    if double_count:
        # Looping over points in `coord_1`
        for i in range(0, Ni):
            # Looping over points in `coord_2`
            for j in range(0, Nj):
                # Calculate the square distances
                sx    = x1[i] - x2[j]
                sy    = y1[i] - y2[j]
                sz    = z1[i] - z2[j]
                lx    = 0.5*(x1[i] + x2[j])
                ly    = 0.5*(y1[i] + y2[j])
                lz    = 0.5*(z1[i] + z2[j])
                l2    = (lx * lx) + (ly * ly) + (lz * lz)
                ll    = sqrt(l2)
                spar  = abs(((sx * lx) + (sy * ly) + (sz * lz)) / ll)
                s2    = (sx * sx) + (sy * sy) + (sz * sz)
                sperp = sqrt(s2 - spar * spar)
                # Criteria for `projected separation`
                if (spar <= pi_max_p) & (sperp > rp_min_p) & (sperp < rp_max_p):
                    # `rp` bin of pair
                    rpbin = (log10(sperp) - logrpmin)/dlogrp
                    # Appending to lists
                    rp_arr.append(rpbin)
                    ith_arr.append(i)
                    jth_arr.append(j)
    else:
        # Looping over points in `coord_1`
        for i in range(Ni):
            # Looping over points in `coord_2`
            for j in range(i+1, Nj):
                # Calculate the square distances
                sx    = x1[i] - x2[j]
                sy    = y1[i] - y2[j]
                sz    = z1[i] - z2[j]
                lx    = 0.5*(x1[i] + x2[j])
                ly    = 0.5*(y1[i] + y2[j])
                lz    = 0.5*(z1[i] + z2[j])
                l2    = (lx * lx) + (ly * ly) + (lz * lz)
                ll    = sqrt(l2)
                spar  = abs(((sx * lx) + (sy * ly) + (sz * lz)) / ll)
                s2    = (sx * sx) + (sy * sy) + (sz * sz)
                sperp = sqrt(s2 - spar * spar)
                # Criteria for `projected separation`
                if (spar <= pi_max_p) & (sperp > rp_min_p) & (sperp < rp_max_p):
                    # `rp` bin of pair
                    rpbin = (log10(sperp) - logrpmin)/dlogrp
                    # Appending to lists
                    rp_arr.append(rpbin)
                    ith_arr.append(i)
                    jth_arr.append(j)
    # Converting to Numpy arrays
    rp_arr     = np.array(rp_arr).astype(int)
    ith_arr    = np.array(ith_arr).astype(int)
    jth_arr    = np.array(jth_arr).astype(int)
    # Combining arrays into a single array
    rp_ith_arr = np.column_stack((rp_arr, ith_arr, jth_arr))

    return rp_ith_arr



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-14
# Last Modified: 2018-05-14
from __future__ import absolute_import, division, print_function
__author__     = ['Victor Calderon']
__copyright__  = ["Copyright 2018 Victor Calderon"]
__email__      = ['victor.calderon@vanderbilt.edu']
__maintainer__ = ['Victor Calderon']
__all__        = [  "Behroozi_relation"]

## Import modules
import numpy as np
from   cosmo_utils.utils             import file_utils as fd
from   cosmo_utils.custom_exceptions import LSSUtils_Error

## Functions

## Retrieves default values for Behroozie et al. (2013)
def _retrieve_Behroozi_default_dict():
    """
    Dictionary of default values of all model parameters set to the
    column 2 values in Table 2 of Behroozi et al. (2013)

    Returns
    --------
    d : `dict`
        Dictionary containing default parameters for the Stellar-Halo
        Mass relation of Behroozi et al. (2013)

    Notes
    ----------
    All calculations are done internally ising the same h=0.7 units as
    in Behroozi ete al. (2010), ['arXiv:1001.0015'] so the parameter values
    here are the same as in Table 2, even though the `mean_log_halo_mass`
    and `mean_stellar_mass` methods use accept and return arguments in
    h=1 units.
    """
    ## Main dictionary
    d = ({'smhm_m0_0': 10.72,
        'smhm_m0_a': 0.59,
        'smhm_m1_0': 12.35,
        'smhm_m1_a': 0.3,
        'smhm_beta_0': 0.43,
        'smhm_beta_a': 0.18,
        'smhm_delta_0': 0.56,
        'smhm_delta_a': 0.18,
        'smhm_gamma_0': 1.54,
        'smhm_gamma_a': 2.52})

    return d

## Behroozi SHMR function
def Behroozi_relation(log_mstar, z=0., return_mhalo_h0=False, mstar_h0=False):
    """
    Returns the halo mass of a central galaxy as a function of its stellar
    mass.

    Parameters
    -----------
    log_mstar : `float` ,`np.ndarray`, or array-like
        Value or array of values of base-10 logarithm of stellar mass
        in h=1 solar mass units.

    z : int, float, `np.ndarray` or array-like
        Redshift of the halo hosting the galaxy. If passing an array,
        it must be of the same length as the input `log_mstar`.

    return_mhalo_h0 : `bool`, optional
        If True, the function returns the halo masses in ``h=1`` units.
        This variable is set to False by default.

    mstar_h0 : `bool`, optional
        If True, the stellar mass in `log_mstar` is converted from ``h=1``
        units to ``h=0.7`` units. This variable is set to False by default.

    Returns
    -----------
    log_halo_mass : float or `np.ndarray`
        Array or float containing 10-base logarithm of halo mass in ``h=1``
        solar mass units.

    Notes
    ----------
    The parameter values in Behroozi+10 were fit to data assuming ``h=0.7``,
    but all halotools inputs are in ``h=1`` units. Thus we will transform
    our input stellar mass to ``h=0.7`` units, evaluate using the
    Behroozi parameters, and then transform back to ``h=1`` units before
    returning the result.
    """
    file_msg = fd.Program_Msg(__file__)
    little_h = 0.7
    ## Checking input parameters
    # `log_mstar`
    mstar_valid_types = (int, float, np.ndarray, list)
    if not (isinstance(log_mstar, mstar_valid_types)):
        msg = '{0} `log_mstar` ({1}) is not a valid type!'.format(
            file_msg, type(log_mstar))
        raise LSSUtils_Error(msg)
    # `z`
    z_valid_types = (int, float, np.ndarray, list)
    if not (isinstance(z, z_valid_types)):
        msg = '{0} `z` ({1}) is not a valid type!'.format(
            file_msg, type(z))
        raise LSSUtils_Error(msg)
    # `return_mhalo_h0`
    return_mhalo_h0_valid_types = (bool)
    if not (isinstance(return_mhalo_h0, return_mhalo_h0_valid_types)):
        msg = '{0} `return_mhalo_h0` ({1}) is not a valid type!'.format(
            file_msg, type(return_mhalo_h0))
        raise LSSUtils_Error(msg)
    # `mstar_h0`
    mstar_h0_valid_types = (bool)
    if not (isinstance(mstar_h0, mstar_h0_valid_types)):
        msg = '{0} `mstar_h0` ({1}) is not a valid type!'.format(
            file_msg, type(mstar_h0))
        raise LSSUtils_Error(msg)
    ##
    ## Behroozi dictionary
    param_dict = _retrieve_Behroozi_default_dict()
    ## Converting from different `h` units
    if mstar_h0:
        mstar = (10**log_mstar)/(little_h**2)
    else:
        mstar = 10.**(log_mstar)
    # Scale factor
    a = 1./(1. + z)
    ##
    ## Behroozi function
    logm0 = param_dict['smhm_m0_0'] + param_dict['smhm_m0_a']*(a - 1)
    m0    = 10.**logm0
    logm1 = param_dict['smhm_m1_0'] + param_dict['smhm_m1_a']*(a - 1)
    beta  = param_dict['smhm_beta_0'] + param_dict['smhm_beta_a']*(a - 1)
    delta = param_dict['smhm_delta_0'] + param_dict['smhm_delta_a']*(a - 1)
    gamma = param_dict['smhm_gamma_0'] + param_dict['smhm_gamma_a']*(a - 1)
    #
    stellar_mass_by_m0 = mstar/m0
    term3_numerator    = (stellar_mass_by_m0)**delta
    term3_denominator  = 1. + (stellar_mass_by_m0)**(-gamma)

    log_halo_mass = logm1 + beta*np.log10(stellar_mass_by_m0)
    log_halo_mass += (term3_numerator/term3_denominator) - 0.5

    # convert back from h=0.7 to h=1 and return the result
    if return_mhalo_h0:
        return np.log10((10.**log_halo_mass)*little_h)
    else:
        return log_halo_mass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-07
# Last Modified: 2018-05-07
from __future__ import print_function, division, absolute_import
__author__     =['Victor Calderon']
__copyright__  =["Copyright 2018 Victor Calderon"]
__email__      =['victor.calderon@vanderbilt.edu']
__maintainer__ =['Victor Calderon']
"""
Set of test functions for the `spherematch` functions
"""

## Import modules
import numpy as np
import pytest
from   cosmo_utils.mock_catalogues import spherematch
from   cosmo_utils.custom_exceptions import LSSUtils_Error

## Functions

## Testing `spherematch` function
def test_spherematch():
    """
    Tests the function 
    `cosmo_utils.mock_catalogues.spherematch.spherematch` for input and 
    output parameters.

    This function tests whether or not a catalogue is matched to 
    its shuffled version.
    """
    ## Initializing random catalogue
    ## Producing set of Right Ascension and Declination arrays
    ra_lim  = (  0, 360.)
    dec_lim = (-90,  90.)
    n_elem  = 1000
    # Right ascension and declination
    ra  = ra_lim [0] + np.random.random_sample(n_elem) * (ra_lim [1]-ra_lim [0])
    dec = dec_lim[0] + np.random.random_sample(n_elem) * (dec_lim[1]-dec_lim[0])
    # 1st sample
    sample1 = np.column_stack((ra, dec))
    # 2nd sample
    sample2 = sample1[:].copy()
    np.random.shuffle(sample2)
    ##
    ## Separating `ra1`, `dec1`, `ra2`, `dec2`
    ra1, dec1 = sample1.T
    ra2, dec2 = sample2.T
    ## Running function
    idx_s1, idx_s2, ds = spherematch.spherematch(   ra1,
                                                    dec1,
                                                    ra2,
                                                    dec2,
                                                    tol=0.01, nnearest=1,
                                                    nthreads=1)
    ## Comparing outcomes
    np.testing.assert_equal(sample1[idx_s1], sample2[idx_s2])



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-04-28
# Last Modified: 2018-04-28
from __future__ import absolute_import, division, print_function
__author__     = ['Victor Calderon']
__copyright__  = ["Copyright 2018 Victor Calderon"]
__email__      = ['victor.calderon@vanderbilt.edu']
__maintainer__ = ['Victor Calderon']
"""
Set of test functions for the `stats_func` functions
"""

## Import modules
import numpy as np
import pytest
from   cosmo_utils.utils import stats_funcs

## Setting seed
seed = np.random.seed(0)
## Functions

##
## Testing `myceil` function
myceil_test = [ (12  , 10, 20.0),
                (13  , 1., 13.0),
                (14.5, 2., 16.0),
                (100 , 10, 100.0),
                (-90 , 10, -90),
                (-89.9, 10, -80),
                (12.05, 0.1, 12.1)]
@pytest.mark.parametrize('x, base, output', myceil_test)
def test_myceil(x, base, output):
    """
    Tests the function `cosmo_utils.utils.stats_funcs.myceil` for input and
    output parameters

    Parameters
    ----------
    x : float
        Number to be approximated to closest number to `base`

    base : float
        Base used to calculate the closest largest number

    output : float
        Closest float number to `x`, i.e. upper-bound float
    """
    myceil_out = stats_funcs.myceil(x, base)

    assert( pytest.approx(myceil_out, base) == output)

##
## Testing `myceil` function
myfloor_test = [(12   , 10 , 10.0 ),
                (13   , 1. , 13.0 ),
                (14.5 , 2. , 12.0 ),
                (99   , 20 , 80.0 ),
                (-90  , 10 , -90  ),
                (-89.9, 10 , -90.0),
                (12.05, 0.1, 12.0 )]
@pytest.mark.parametrize('x, base, output', myfloor_test)
def test_myfloor(x, base, output):
    """
    Tests the function `cosmo_utils.utils.stats_funcs.myceil` for input and
    output parameters

    Parameters
    ----------
    x : float
        Number to be approximated to closest number to `base`

    base : float
        Base used to calculate the closest largest number

    output : float
        Closest float number to `x`, i.e. upper-bound float
    """
    myceil_out = stats_funcs.myfloor(x, base)

    assert( pytest.approx(myceil_out, base) == output)

##
## Testing `Bins_array_create` function
bins_arr_test = [   ([1, 2, 3, 4, 5], 2, [0., 2., 4, 6.]),
                    ([1.2, 2.5, 3, 4], 0.8, [0.8, 1.6, 2.4, 3.2, 4.0]),
                    ([2.25, 3.4, 4.2, 4.3, 2.6], 2, [2., 4, 6.]),
                    ([-6, 2, -8.5, 3.], 3, [-9, -6, -3, 0, 3]),
                    ([1, 2, 3.5, 4], 2, [0., 2, 4.]),
                    ([-100, 200, 0. -12.5], 50., [-100, -50, 0., 50, 100, 150, 200])]
@pytest.mark.parametrize('arr, base, output', bins_arr_test)
def test_Bins_array_create(arr, base, output):
    """
    Tests the function `cosmo_utils.utils.stats_funcs.Bins_array_create` for
    input and output parameters

    Parameters
    -----------
    arr : array_like
        Array of of numbers or floats

    base : int or float, optional
        Interval used to create the evenly-spaced array of elements

    output : `numpy.ndarray`
        Array of elements separated in intervals of `base`
    """
    output = np.array(output)
    # Output from `Bins_array_create` function
    bins_arr = stats_funcs.Bins_array_create(arr, base)
    # Checking that array are equal
    assert(np.allclose(output, bins_arr))

##
## Testing `sigma_calcs`
sigma_calcs_test = [    (1, 100 , (1,100)),
                        (2, 1000, (2,100))]
@pytest.mark.parametrize('return_mean, out_type',[(True, tuple),(False, dict)])
@pytest.mark.parametrize('type_sigma',['std'])
@pytest.mark.parametrize('ndim, nelem, output', sigma_calcs_test)
def test_sigma_calcs_shape(ndim, nelem, output, type_sigma,
    return_mean, out_type):
    """
    Tests the function `cosmo_utils.utils.stats_funcs.sigma_calcs` for
    input and output parameters

    Parameters
    -----------
    ndim : int
        Dimesions of the array being tested

    nelem : int
        Number of elements in each sub-array

    output : tuple
        Tuple of elements

    type_sigma : {'std', 'perc'} str
        Type of statistics to use for the different sigmas/St. Dev.

    return_mean : `bool`
        If true, it returns `mean`, `std` and `sigmas`.
        If False, it only return `sigmas`.

    out_type : {list, dict} type object
        Type object for the given `return_mean` option
    """
    ## Unpacking input parameters
    ndim_out, len_out = output
    ## Generating dataset
    arr_test = np.random.random((ndim, nelem))
    ## Determining output from `sigma_calcs`
    dict_test      = stats_funcs.sigma_calcs(arr_test, type_sigma=type_sigma,
                        return_mean_std=False)
    dict_test_keys = dict_test.keys()
    ## Testing dimensions and lengths
    for ii in dict_test_keys:
        # Dimensions
        dict_ii_ndim = dict_test[ii].ndim
        assert(dict_ii_ndim == 2)
        # Lengths
        dict_ii_len  = len(dict_test[ii])
        assert(dict_ii_len == 2)
    ##
    ## Testing datatypes
    dict_test_all = stats_funcs.sigma_calcs(arr_test, type_sigma=type_sigma,
                        return_mean_std=return_mean)
    assert(type(dict_test_all) == out_type)
    # Testing types for when `return_mean` == 'True'
    if return_mean:
        assert(type(dict_test_all[0]) == dict)
        assert(type(dict_test_all[1]) == np.ndarray)
        assert(type(dict_test_all[2]) == np.ndarray)

##
## Testing `Stats_one_arr`
test_return_perc_arr     = [(True , 1),
                            (False, 0)]
arr_digit_expected_arr   = [('n', 4),
                            ('y', 6),
                            ('o', 2)]
test_stats_one_nelem_arr = [100, 1000, 10000, 50, 20]
@pytest.mark.parametrize('return_perc, return_perc_val', test_return_perc_arr)
@pytest.mark.parametrize('arr_digit, expected', arr_digit_expected_arr)
@pytest.mark.parametrize('nelem', test_stats_one_nelem_arr)
def test_Stats_one_arr_outputs(return_perc, return_perc_val, arr_digit,
    expected, nelem, bins=10, base=10):
    """
    Tests the function `cosmo_utils.utils.stats_funcs.Stats_one_arr` for
    input and output parameters

    Parameters
    ----------
    return_perc : `bool`, optional
        If true, it also returns the `percentiles` of the data.
        Last item in the return list.
        This variable is set to False by default.

    return_perc_val : int
        Value to add to `expected`, given if returning the percentiles
        of the bins or not.

    arr_digit : {'n', 'y', 'o'} str, optional
        Option for which elements to return.
        - 'n' : Returns `x1_stat`, `y1_stat`, `y1_std`, `y1_std_err`
        - 'y' : Returns `x1_stat`, `y1_stat`, `y1_std`, `y1_std_err`,
                        `x_bins_data`, `y_bins_data`
        - 'o' : Returns `x_bins_data`, `y_bins_data`

    expected : int
        Expected number of elements in the tuple, based on the input
        arguments

    nelem : int
        Size of the array. It will test if all of the elements are
        being returned correctly
    """
    ## Creating random arrays
    x = np.random.random(nelem) * 50.
    y = np.random.random(nelem) * 100.
    ##
    ## Running function
    output = stats_funcs.Stats_one_arr(x, y, base=base,
                arr_digit=arr_digit, statfunc=np.nanmean,
                return_perc=return_perc)
    ##
    ## Testing number of elements returned
    output_expected = expected + return_perc_val
    assert(len(output) == output_expected)
    ##
    ## Checking type of all of the returned items
    if return_perc:
        for ii in output[:-1]:
            assert(isinstance(ii, np.ndarray))
        assert(isinstance(output[-1], dict))
    else:
        for ii in output:
            assert(isinstance(ii, np.ndarray))
    ## Number of elements in array
    (   x_bins_data,
        y_bins_data) = stats_funcs.Stats_one_arr(x, y, base=base,
                        arr_digit='o', statfunc=np.nanmean, arr_len=0,
                        return_perc=False)
    assert(len(np.concatenate(x_bins_data).ravel()) == nelem)
    assert(len(np.concatenate(y_bins_data).ravel()) == nelem)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-08
# Last Modified: 2018-05-08
from __future__ import print_function, division, absolute_import
__author__     =['Victor Calderon']
__copyright__  =["Copyright 2018 Victor Calderon"]
__email__      =['victor.calderon@vanderbilt.edu']
__maintainer__ =['Victor Calderon']
"""
Set of test functions for the `abundance_matching` functions
"""

## Import modules
import numpy as np
import pytest
from   cosmo_utils.mock_catalogues import abundance_matching
from   cosmo_utils.custom_exceptions import LSSUtils_Error

## Functions

## Test `reversed_arrays` - Inputs
reversed_arrays_input_arr = [   ([1,2,3,4], [10,15, 16, 20], True),
                                ([-5,-6,-7,-8], [-10,-11,-12,-20], False),
                                ([10,20,30],[-20,0, 20], True)]
@pytest.mark.parametrize('x, y, expected', reversed_arrays_input_arr)
def test_reversed_arrays(x, y, expected):
    """
    Tests the function 
    `cosmo_utils.mock_catalogues.abundance_matching.reversed_arrays`
    for input and output parameters.

    This function tests whether or not the 2 arrays are monotonically
    increasing with each other or not.

    Parameters
    ----------
    x : `numpy.ndarray`
        Array containing the 1st set of values

    y : `numpy.ndarray`
        Array containing the 2nd set of values.

    expected : `bool`
        If True, `x` increases monotonically with increasing `y`.
        If False, `x` decreases monotonically with increasing `y`.
    """
    ## Running function
    output = abundance_matching.reversed_arrays(x, y)
    ## Checking against expected value
    assert(output == expected)





# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2018-05-02 16:56:40
from __future__ import unicode_literals
import datetime

version = "3.0.1"
githash = "09bbe9583e70ed1804f1051a3aaca1640d08c6f8"


major = 3
minor = 0
bugfix = 1

release = True
timestamp = datetime.datetime(2018, 5, 2, 16, 56, 40)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-07
# Last Modified: 2018-05-07
from __future__ import absolute_import, division, print_function
__author__     = ['Victor Calderon']
__copyright__  = ["Copyright 2018 Victor Calderon"]
__email__      = ['victor.calderon@vanderbilt.edu']
__maintainer__ = ['Victor Calderon']
"""
Set of test functions for the `geometry` functions
"""

## Import modules
import numpy as np
import pandas as pd
import pytest
from   cosmo_utils.utils import geometry
from   cosmo_utils.utils import file_utils as fd
from   cosmo_utils.custom_exceptions import LSSUtils_Error


## Setting seed
seed = np.random.seed(0)

## Functions

## Testing function `flip_angles`
flip_angles_test_arr = [    (-50, 310.0),
                            (130.5, 130.5),
                            (-0.5, 359.5),
                            ([-10., -20.5, 60.], np.array([350.0, 339.5, 60.0])),
                            (np.array([12., -10]), np.array([12.0, 350.0]))]
@pytest.mark.parametrize('input_ang, output_ang', flip_angles_test_arr)
def test_flip_angles(input_ang, output_ang):
    """
    Tests the function `cosmo_utils.utils.geometry.flip_angles` for input and
    output parameters

    Parameters
    -----------
    input_ang : float, int, list, `numpy.ndarray`
        Input argument to evaluate

    output_ang : float, int, list, `numpy.ndarray`
        Result to test against. It is the expected result from `flip_angles`
        function.
    """
    ## Checking that outputs from function are the ones to be expected
    output_func = geometry.flip_angles(input_ang)
    # Checking agains 
    np.testing.assert_allclose(output_func, output_ang)

## Testing function Ang_Distance
def test_Ang_Distance_comparison_astropy():
    """
    Tests the function `cosmo_utils.utils.geometry.flip_angles` for input and
    output parameters.

    This method compares the values obtained from the `haversine` to those
    using the `astropy` method.
    """
    ## Producing set of Right Ascension and Declination arrays
    ra_lim  = (0, 360.)
    dec_lim = (-90, 90.)
    for ii in range(1, 1000, 10):
        ## Random array for RA and Dec
        ra1  = ra_lim[0]  + np.random.random_sample(ii) * (ra_lim [1]-ra_lim [0])
        dec1 = dec_lim[0] + np.random.random_sample(ii) * (dec_lim[1]-dec_lim[0])
        ra2  = ra_lim[0]  + np.random.random_sample(ii) * (ra_lim [1]-ra_lim [0])
        dec2 = dec_lim[0] + np.random.random_sample(ii) * (dec_lim[1]-dec_lim[0])
        ## Testing outputs from different methods
        out_haversine = geometry.Ang_Distance(ra1, ra2, dec1, dec2,
            method='haversine', unit='deg')
        out_astropy   = geometry.Ang_Distance(ra1, ra2, dec1, dec2,
            method='astropy', unit='deg')
        ## Checking that arrays are the same or similar
        np.testing.assert_allclose(out_haversine, out_astropy)

## Testing `Ang_Distance` for errors - Units
Ang_Distance_test_unit_arr = ['deg2', 'nan']
@pytest.mark.parametrize('unit', Ang_Distance_test_unit_arr)
def test_Ang_Distance_unit_errors(unit):
    """
    Tests the function `cosmo_utils.utils.geometry.flip_angles` for input and
    output parameters.

    This function makes sure that errors are raised whenever a wrong
    input is given.

    Parameters
    -----------
    unit : {'dec','rad'} str, optional
        Unit of `ra1`, `ra2`, `dec1`, and `dec2`.
        This will also determine the final unit that outputs this function.

    method : {'haversine', 'astropy'} str, optional
        Method to use in order to calculate angular separation.
        This variable is to by default to the `haversine` method.
        If `astropy`, it will use the astropy framework to determine the
        angular separation.
    """
    ## Producing set of Right Ascension and Declination arrays
    ra_lim  = (  0, 360.)
    dec_lim = (-90,  90.)
    for ii in range(1, 1000, 10):
        ## Random array for RA and Dec
        ra1  = ra_lim [0] + np.random.random_sample(ii) * (ra_lim [1]-ra_lim [0])
        dec1 = dec_lim[0] + np.random.random_sample(ii) * (dec_lim[1]-dec_lim[0])
        ra2  = ra_lim [0] + np.random.random_sample(ii) * (ra_lim [1]-ra_lim [0])
        dec2 = dec_lim[0] + np.random.random_sample(ii) * (dec_lim[1]-dec_lim[0])
        ## Testing outputs from different methods
        # Haversine method
        with pytest.raises(LSSUtils_Error):
            out_astropy   = geometry.Ang_Distance(ra1, ra2, dec1, dec2,
                method='astropy', unit=unit)

## Testing `Ang_Distance` for errors - Method
Ang_Distance_test_method_arr = ['meter', 'NotMethod']
@pytest.mark.parametrize('method', Ang_Distance_test_method_arr)
def test_Ang_Distance_method_errors(method):
    """
    Tests the function `cosmo_utils.utils.geometry.flip_angles` for input and
    output parameters.

    This function makes sure that errors are raised whenever a wrong
    input is given.

    Parameters
    -----------
    unit : {'dec','rad'} str, optional
        Unit of `ra1`, `ra2`, `dec1`, and `dec2`.
        This will also determine the final unit that outputs this function.

    method : {'haversine', 'astropy'} str, optional
        Method to use in order to calculate angular separation.
        This variable is to by default to the `haversine` method.
        If `astropy`, it will use the astropy framework to determine the
        angular separation.
    """
    ## Producing set of Right Ascension and Declination arrays
    ra_lim  = (  0, 360.)
    dec_lim = (-90,  90.)
    for ii in range(1, 1000, 10):
        ## Random array for RA and Dec
        ra1  = ra_lim [0] + np.random.random_sample(ii) * (ra_lim [1]-ra_lim [0])
        dec1 = dec_lim[0] + np.random.random_sample(ii) * (dec_lim[1]-dec_lim[0])
        ra2  = ra_lim [0] + np.random.random_sample(ii) * (ra_lim [1]-ra_lim [0])
        dec2 = dec_lim[0] + np.random.random_sample(ii) * (dec_lim[1]-dec_lim[0])
        ## Testing outputs from different methods
        # Haversine method
        with pytest.raises(LSSUtils_Error):
            out_astropy = geometry.Ang_Distance(ra1, ra2, dec1, dec2,
                method=method, unit='deg')

## Testing `Coord_Transformation` for type of what it returns
Coord_test_return = [   (True, dict),
                        (False, pd.DataFrame)]
@pytest.mark.parametrize('return_dict, return_type', Coord_test_return)
def test_Coord_Transformation_types(return_dict, return_type):
    """
    Tests the function `cosmo_utils.utils.geometry.Coord_Transformation`
    for input and output parameters.

    This function makes sure that errors are raised whenever a wrong
    input is given.
    """
    ## Expected keys in dict/DataFrame
    expected_keys = np.sort(['ra', 'dec', 'dist', 'x', 'y', 'z'])
    ## Producing set of Right Ascension and Declination arrays
    ra_lim   = (  0, 360.)
    dec_lim  = (-90,  90.)
    dist_lim = (1e-2, 100.)
    for ii in range(1000, 10000, 100):
        ## Random array for RA, DEC, and DIST
        # Centre
        ra_cen   = ra_lim  [0] + (ra_lim  [1]-ra_lim  [0]) * np.random.random_sample()
        dec_cen  = dec_lim [0] + (dec_lim [1]-dec_lim [0]) * np.random.random_sample()
        dist_cen = dist_lim[0] + (dist_lim[1]-dist_lim[0]) * np.random.random_sample()
        # Rest of elements
        ra   = ra_lim  [0] + (ra_lim  [1]-ra_lim  [0]) * np.random.random_sample(ii)
        dec  = dec_lim [0] + (dec_lim [1]-dec_lim [0]) * np.random.random_sample(ii)
        dist = dist_lim[0] + (dist_lim[1]-dist_lim[0]) * np.random.random_sample(ii)
        ##
        ## Generating outputs
        output = geometry.Coord_Transformation( ra,
                                                dec,
                                                dist,
                                                ra_cen,
                                                dec_cen,
                                                dist_cen,
                                                trans_opt=1,
                                                return_dict=return_dict)
        ## Checking return type
        assert(isinstance(output, return_type))
        ## Checking data returned
        if return_dict:
            output_keys = np.sort(list(output.keys()))
        else:
            output_keys = np.sort(output.columns.values)
        assert(np.array_equal(expected_keys, output_keys))
        ##
        ## Checking sizes
        for elem in expected_keys:
            assert(len(output[elem]) == ii)

## Testing `Coord_Transformation` for errors - Units
Coord_test_unit_errors = ['nan', 'NotDegree']
@pytest.mark.parametrize('unit', Coord_test_unit_errors)
def test_Coord_Transformation_errors(unit):
    """
    Tests the function `cosmo_utils.utils.geometry.Coord_Transformation`
    for input and output parameters.

    This function makes sure that errors are raised whenever a wrong
    input is given.

    Parameters
    ----------
    unit : {'dec','rad'} str, optional
        Unit of `ra1`, `ra2`, `dec1`, and `dec2`.
        This will also determine the final unit that outputs this function.
        This variable is set to `deg` by default.


    """
    ## Producing set of Right Ascension and Declination arrays
    ra_lim   = (  0, 360.)
    dec_lim  = (-90,  90.)
    dist_lim = (1e-2, 100.)
    for ii in range(1000, 10000, 100):
        ## Random array for RA, DEC, and DIST
        # Centre
        ra_cen   = ra_lim  [0] + (ra_lim  [1]-ra_lim  [0]) * np.random.random_sample()
        dec_cen  = dec_lim [0] + (dec_lim [1]-dec_lim [0]) * np.random.random_sample()
        dist_cen = dist_lim[0] + (dist_lim[1]-dist_lim[0]) * np.random.random_sample()
        # Rest of elements
        ra   = ra_lim  [0] + (ra_lim  [1]-ra_lim  [0]) * np.random.random_sample(ii)
        dec  = dec_lim [0] + (dec_lim [1]-dec_lim [0]) * np.random.random_sample(ii)
        dist = dist_lim[0] + (dist_lim[1]-dist_lim[0]) * np.random.random_sample(ii)
        ##
        ## Generating outputs
        with pytest.raises(LSSUtils_Error):
            output = geometry.Coord_Transformation( ra,
                                                    dec,
                                                    dist,
                                                    ra_cen,
                                                    dec_cen,
                                                    dist_cen,
                                                    trans_opt=1,
                                                    return_dict=True,
                                                    unit=unit)

## Testing `Coord_Transformation` for errors - trans_opt
Coord_test_trans_opt_errors = list(range(5, 10))
@pytest.mark.parametrize('trans_opt', Coord_test_trans_opt_errors)
def test_Coord_Transformation_errors(trans_opt):
    """
    Tests the function `cosmo_utils.utils.geometry.Coord_Transformation`
    for input and output parameters.

    This function makes sure that errors are raised whenever a wrong
    input is given.

    Parameters
    ----------
    trans_opt : {1, 2, 3, 4} int, optional
        Option for cartesian translation/transformation for elements.
        This variable ist set to `4` by default.

        Options:
            - 1 : No translation involved
            - 2 : Translation to the center point.
            - 3 : Translation `and` rotation to the center point.
            - 4 : Translation and 2 rotaitons about the center point

    """
    ## Producing set of Right Ascension and Declination arrays
    ra_lim   = (  0, 360.)
    dec_lim  = (-90,  90.)
    dist_lim = (1e-2, 100.)
    for ii in range(1000, 10000, 100):
        ## Random array for RA, DEC, and DIST
        # Centre
        ra_cen   = ra_lim  [0] + (ra_lim  [1]-ra_lim  [0]) * np.random.random_sample()
        dec_cen  = dec_lim [0] + (dec_lim [1]-dec_lim [0]) * np.random.random_sample()
        dist_cen = dist_lim[0] + (dist_lim[1]-dist_lim[0]) * np.random.random_sample()
        # Rest of elements
        ra   = ra_lim  [0] + (ra_lim  [1]-ra_lim  [0]) * np.random.random_sample(ii)
        dec  = dec_lim [0] + (dec_lim [1]-dec_lim [0]) * np.random.random_sample(ii)
        dist = dist_lim[0] + (dist_lim[1]-dist_lim[0]) * np.random.random_sample(ii)
        ##
        ## Generating outputs
        with pytest.raises(LSSUtils_Error):
            output = geometry.Coord_Transformation( ra,
                                                    dec,
                                                    dist,
                                                    ra_cen,
                                                    dec_cen,
                                                    dist_cen,
                                                    trans_opt=trans_opt,
                                                    return_dict=True,
                                                    unit='deg')

## Testing `Coord_Transformation` for errors - ra_cen type
Coord_test_ra_cen_errors = ['1', 'String']
@pytest.mark.parametrize('ra_cen', Coord_test_ra_cen_errors)
def test_Coord_Transformation_errors(ra_cen):
    """
    Tests the function `cosmo_utils.utils.geometry.Coord_Transformation`
    for input and output parameters.

    This function makes sure that errors are raised whenever a wrong
    input is given.

    Parameters
    ----------
    ra_cen : float, int
        Right Ascension, declination, and distance for the center of
        the coordinates. These correspond to where the corodinates
        `ra`, `dec`, and `dist` will be centered.

    """
    ## Producing set of Right Ascension and Declination arrays
    ra_lim   = (  0, 360.)
    dec_lim  = (-90,  90.)
    dist_lim = (1e-2, 100.)
    for ii in range(1000, 10000, 100):
        ## Random array for RA, DEC, and DIST
        # Centre
        dec_cen  = dec_lim [0] + (dec_lim [1]-dec_lim [0]) * np.random.random_sample()
        dist_cen = dist_lim[0] + (dist_lim[1]-dist_lim[0]) * np.random.random_sample()
        # Rest of elements
        ra   = ra_lim  [0] + (ra_lim  [1]-ra_lim  [0]) * np.random.random_sample(ii)
        dec  = dec_lim [0] + (dec_lim [1]-dec_lim [0]) * np.random.random_sample(ii)
        dist = dist_lim[0] + (dist_lim[1]-dist_lim[0]) * np.random.random_sample(ii)
        ##
        ## Generating outputs
        with pytest.raises(LSSUtils_Error):
            output = geometry.Coord_Transformation( ra,
                                                    dec,
                                                    dist,
                                                    ra_cen,
                                                    dec_cen,
                                                    dist_cen,
                                                    trans_opt=1,
                                                    return_dict=True,
                                                    unit='deg')



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-08
# Last Modified: 2018-05-08
from __future__ import print_function, division, absolute_import
__author__     =['Victor Calderon']
__copyright__  =["Copyright 2018 Victor Calderon"]
__email__      =['victor.calderon@vanderbilt.edu']
__maintainer__ =['Victor Calderon']
"""
Set of test functions for the `catls_utils` functions
"""

## Import modules
import numpy as np
import pytest
from   cosmo_utils.mock_catalogues import catls_utils
from   cosmo_utils.custom_exceptions import LSSUtils_Error

## Functions

## Test `catl_keys` function - Types
catl_keys_types_arr = [     ('data' , 'list', 3, list),
                            ('data' , 'dict', 3, dict),
                            ('mocks', 'list', 3, list),
                            ('mocks', 'dict', 3, dict) ]
@pytest.mark.parametrize('catl_kind, return_type, nelem, expected',
    catl_keys_types_arr)
def test_catl_keys_types_nelem(catl_kind, return_type, nelem, expected):
    """
    Tests the function:
        cosmo_utils.mock_catalogues.catls_utils.catl_keys` for input and 
    output variables.

    It verifies the `type` of the output returned by the function.

    Parameters
    -----------
    catl_kind : {'data', 'mocks'} str
        Type of catalogue to use. This variable is set to `data` by default.

        Options:
            - `data` : catalogues come from SDSS `real` catalogue
            - `mocks` : catalogue come from SDSS `mock` catalogues

    return_type : {'list', 'dict'} str
        Type of output to the be returned. This variable is set to `list`
        by default.

        Options:
            - 'list' : Returns the values as part of a list
            - 'dict' : Returns the values as part of a python dictionary

    nelem : int
        Expected number of elements inside the object returned by the function.

    expected : str
        Expected type of element from the `catl_keys` function
    """
    ## Constants
    perf_opt = False
    ## Running element
    output = catls_utils.catl_keys(catl_kind, return_type=return_type,
        perf_opt=perf_opt)
    ## Comparing against `expected` value - Type
    assert(isinstance(output, expected))
    ## Checking number of elements returned
    if isinstance(output, list):
        assert(len(output) == nelem)
    elif isinstance(output, dict):
        assert(len(output.keys()) == nelem)

## Test `catl_keys` function - Outputs
catl_keys_return_arr = [ 'list' , 'dict']
catl_keys_output_arr = [('data' , False, ['M_h', 'groupid', 'galtype']),
                        ('data' , False, ['M_h', 'groupid', 'galtype']),
                        ('mocks', False, ['M_group', 'groupid', 'g_galtype']),
                        ('mocks', True, ['M_h', 'haloid', 'galtype'])]
@pytest.mark.parametrize('return_type', catl_keys_return_arr)
@pytest.mark.parametrize('catl_kind, perf_opt, expected', catl_keys_output_arr)
def test_catl_keys_outputs(catl_kind, perf_opt, return_type, expected):
    """
    Tests the function:
        cosmo_utils.mock_catalogues.catls_utils.catl_keys` for input and 
    output variables.

    It verifies the output returned by the function.

    Parameters
    -----------
    catl_kind : {'data', 'mocks'} str
        Type of catalogue to use. This variable is set to `data` by default.

        Options:
            - `data` : catalogues come from SDSS `real` catalogue
            - `mocks` : catalogue come from SDSS `mock` catalogues

    perf_opt : `bool`, optional
        Option for using a `perfect` mock catalogue.

    return_type : {'list', 'dict'} str
        Type of output to the be returned. This variable is set to `list`
        by default.

        Options:
            - 'list' : Returns the values as part of a list
            - 'dict' : Returns the values as part of a python dictionary

    expected : str
        Expected type of element from the `catl_keys` function
    """
    ## Running element
    output = catls_utils.catl_keys(catl_kind, perf_opt=perf_opt, 
        return_type=return_type)
    ## Comparing against `expected` value - Output
    if isinstance(output, list):
        np.testing.assert_equal(output, expected)
    elif isinstance(output, dict):
        out_keys = ['gm_key', 'id_key', 'galtype_key']
        out_vals = [output[xx] for xx in out_keys]
        np.testing.assert_equal(out_vals, expected)

## Test `catl_keys` function - Errors - `catl_kind`
catl_keys_catl_kind_arr = [ 'data1', 'mocks1', 'NoMethod']
catl_keys_catl_perf_arr = [ True, False]
catl_keys_return_arr    = [ 'list' , 'dict']
@pytest.mark.parametrize('catl_kind', catl_keys_catl_kind_arr)
@pytest.mark.parametrize('return_type', catl_keys_return_arr)
@pytest.mark.parametrize('perf_opt', catl_keys_catl_perf_arr)
def test_catl_keys_catl_kind_errors(catl_kind, perf_opt, return_type):
    """
    Tests the function:
        cosmo_utils.mock_catalogues.catls_utils.catl_keys` for input and 
    output variables.

    It verifies if errors are raised when `catl_kind` is incorrect

    Parameters
    -----------
    catl_kind : {'data', 'mocks'} str
        Type of catalogue to use. This variable is set to `data` by default.

        Options:
            - `data` : catalogues come from SDSS `real` catalogue
            - `mocks` : catalogue come from SDSS `mock` catalogues
    """
    ## Running function
    with pytest.raises(LSSUtils_Error):
        output = catls_utils.catl_keys(catl_kind, perf_opt=perf_opt,
            return_type=return_type)

## Test `catl_keys` function - Errors - `return_type`
catl_keys_catl_kind_arr = ['data', 'mocks']
catl_keys_catl_perf_arr = [True, False]
catl_keys_return_arr    = [ 'list_no' , 'dict1', 'NoMethod']
@pytest.mark.parametrize('catl_kind', catl_keys_catl_kind_arr)
@pytest.mark.parametrize('return_type', catl_keys_return_arr)
@pytest.mark.parametrize('perf_opt', catl_keys_catl_perf_arr)
def test_catl_keys_catl_kind_errors(catl_kind, perf_opt, return_type):
    """
    Tests the function:
        cosmo_utils.mock_catalogues.catls_utils.catl_keys` for input and 
    output variables.

    It verifies if errors are raised when `catl_kind` is incorrect

    Parameters
    -----------
    catl_kind : {'data', 'mocks'} str
        Type of catalogue to use. This variable is set to `data` by default.

        Options:
            - `data` : catalogues come from SDSS `real` catalogue
            - `mocks` : catalogue come from SDSS `mock` catalogues
    """
    ## Running function
    with pytest.raises(LSSUtils_Error):
        output = catls_utils.catl_keys(catl_kind, perf_opt=perf_opt,
            return_type=return_type)

## Test `catl_keys` function - Errors - `return_type`
catl_keys_catl_kind_arr = ['data', 'mocks']
catl_keys_catl_perf_arr = [ 'NotBoolean', 1, 'mark', 1.2]
catl_keys_return_arr    = [ 'list' , 'dict']
@pytest.mark.parametrize('catl_kind', catl_keys_catl_kind_arr)
@pytest.mark.parametrize('return_type', catl_keys_return_arr)
@pytest.mark.parametrize('perf_opt', catl_keys_catl_perf_arr)
def test_catl_keys_catl_kind_errors(catl_kind, perf_opt, return_type):
    """
    Tests the function:
        cosmo_utils.mock_catalogues.catls_utils.catl_keys` for input and 
    output variables.

    It verifies if errors are raised when `catl_kind` is incorrect

    Parameters
    -----------
    catl_kind : {'data', 'mocks'} str
        Type of catalogue to use. This variable is set to `data` by default.

        Options:
            - `data` : catalogues come from SDSS `real` catalogue
            - `mocks` : catalogue come from SDSS `mock` catalogues
    """
    ## Running function
    with pytest.raises(LSSUtils_Error):
        output = catls_utils.catl_keys(catl_kind, perf_opt=perf_opt,
            return_type=return_type)

#########-------------------------------------------------------------#########

## Test `catl_keys_prop` function - Types
catl_keys_prop_info_arr  = ['members', 'groups']
catl_keys_prop_types_arr = [('data' , 'list', 2, list),
                            ('data' , 'dict', 2, dict),
                            ('mocks', 'list', 2, list),
                            ('mocks', 'dict', 2, dict) ]
@pytest.mark.parametrize('catl_info', catl_keys_prop_info_arr)
@pytest.mark.parametrize('catl_kind, return_type, nelem, expected',
    catl_keys_prop_types_arr)
def test_catl_keys_prop_types_nelem(catl_kind, catl_info, return_type, nelem,
    expected):
    """
    Tests the function:
        cosmo_utils.mock_catalogues.catls_utils.catl_keys_prop` for input and 
    output variables.

    It verifies the `type` of the output returned by the function.

    Parameters
    -----------
    catl_kind : {'data', 'mocks'} str
        Type of catalogue to use. This variable is set to `data` by default.

        Options:
            - `data` : catalogues come from SDSS `real` catalogue
            - `mocks` : catalogue come from SDSS `mock` catalogues

    catl_info : {'members', 'groups'} str, optional
        Option for which kind of catalogues to use.

    return_type : {'list', 'dict'} str
        Type of output to the be returned. This variable is set to `list`
        by default.

        Options:
            - 'list' : Returns the values as part of a list
            - 'dict' : Returns the values as part of a python dictionary

    nelem : int
        Expected number of elements inside the object returned by the function.

    expected : str
        Expected type of element from the `catl_keys_prop` function
    """
    ## Running element
    output = catls_utils.catl_keys_prop(catl_kind, catl_info=catl_info,
        return_type=return_type)
    ## Comparing against `expected` value - Type
    assert(isinstance(output, expected))
    ## Checking number of elements returned
    if isinstance(output, list):
        assert(len(output) == nelem)
    elif isinstance(output, dict):
        assert(len(output.keys()) == nelem)

## Test `catl_keys_prop` function - Outputs
catl_keys_prop_return_arr = [ 'list' , 'dict']
catl_keys_prop_output_arr = [('data' , 'members', ['logssfr'    , 'logMstar_JHU']),
                        ('data' , 'groups' , ['logssfr_tot', 'logMstar_tot']),
                        ('mocks', 'members', ['logssfr'    , 'logMstar']),
                        ('mocks', 'groups' , ['logssfr'    , 'logMstar'])]
@pytest.mark.parametrize('return_type', catl_keys_prop_return_arr)
@pytest.mark.parametrize('catl_kind, catl_info, expected', catl_keys_prop_output_arr)
def test_catl_keys_prop_outputs(catl_kind, catl_info, return_type, expected):
    """
    Tests the function:
        cosmo_utils.mock_catalogues.catls_utils.catl_keys_prop` for input and 
    output variables.

    It verifies the output returned by the function.

    Parameters
    -----------
    catl_kind : {'data', 'mocks'} str
        Type of catalogue to use. This variable is set to `data` by default.

        Options:
            - `data` : catalogues come from SDSS `real` catalogue
            - `mocks` : catalogue come from SDSS `mock` catalogues

    catl_info : {'members', 'groups'} str, optional
        Option for which kind of catalogues to use.

    return_type : {'list', 'dict'} str
        Type of output to the be returned. This variable is set to `list`
        by default.

        Options:
            - 'list' : Returns the values as part of a list
            - 'dict' : Returns the values as part of a python dictionary

    expected : str
        Expected type of element from the `catl_keys_prop` function
    """
    ## Running element
    output = catls_utils.catl_keys_prop(catl_kind, catl_info=catl_info, 
        return_type=return_type)
    ## Comparing against `expected` value - Output
    if isinstance(output, list):
        np.testing.assert_equal(output, expected)
    elif isinstance(output, dict):
        out_keys = ['logssfr_key', 'logmstar_key']
        out_vals = [output[xx] for xx in out_keys]
        np.testing.assert_equal(out_vals, expected)

## Test `catl_keys_prop` function - Errors - `catl_kind`
catl_keys_prop_catl_kind_arr = [ 'data1', 'mocks1', 'NoMethod']
catl_keys_prop_return_arr    = [ 'list' , 'dict']
catl_keys_prop_catl_info_arr = [ 'members', 'groups']
@pytest.mark.parametrize('catl_kind', catl_keys_prop_catl_kind_arr)
@pytest.mark.parametrize('return_type', catl_keys_prop_return_arr)
@pytest.mark.parametrize('catl_info', catl_keys_prop_catl_info_arr)
def test_catl_keys_prop_catl_kind_errors(catl_kind, catl_info, return_type):
    """
    Tests the function:
        cosmo_utils.mock_catalogues.catls_utils.catl_keys_prop` for input and 
    output variables.

    It verifies if errors are raised when `catl_kind` is incorrect

    Parameters
    -----------
    catl_kind : {'data', 'mocks'} str
        Type of catalogue to use. This variable is set to `data` by default.

        Options:
            - `data` : catalogues come from SDSS `real` catalogue
            - `mocks` : catalogue come from SDSS `mock` catalogues
    """
    ## Running function
    with pytest.raises(LSSUtils_Error):
        output = catls_utils.catl_keys_prop(catl_kind, catl_info=catl_info,
            return_type=return_type)

## Test `catl_keys_prop` function - Errors - `catl_info`
catl_keys_prop_catl_kind_arr = [ 'data', 'mocks']
catl_keys_prop_return_arr    = [ 'list' , 'dict']
catl_keys_prop_catl_info_arr = [ 'members_no', 'groups_Invalid', 1, 1.2]
@pytest.mark.parametrize('catl_kind', catl_keys_prop_catl_kind_arr)
@pytest.mark.parametrize('return_type', catl_keys_prop_return_arr)
@pytest.mark.parametrize('catl_info', catl_keys_prop_catl_info_arr)
def test_catl_keys_prop_catl_info_errors(catl_kind, catl_info, return_type):
    """
    Tests the function:
        cosmo_utils.mock_catalogues.catls_utils.catl_keys_prop` for input and 
    output variables.

    It verifies if errors are raised when `catl_info` is incorrect

    Parameters
    -----------
    catl_kind : {'data', 'mocks'} str
        Type of catalogue to use. This variable is set to `data` by default.

        Options:
            - `data` : catalogues come from SDSS `real` catalogue
            - `mocks` : catalogue come from SDSS `mock` catalogues
    """
    ## Running function
    with pytest.raises(LSSUtils_Error):
        output = catls_utils.catl_keys_prop(catl_kind, catl_info=catl_info,
            return_type=return_type)

## Test `catl_keys_prop` function - Errors - `return_type`
catl_keys_prop_catl_kind_arr = [ 'data', 'mocks']
catl_keys_prop_return_arr    = [ 'list_no' , 'dict1', 'NoMethod']
catl_keys_prop_catl_info_arr = [ 'members', 'groups']
@pytest.mark.parametrize('catl_kind', catl_keys_prop_catl_kind_arr)
@pytest.mark.parametrize('return_type', catl_keys_prop_return_arr)
@pytest.mark.parametrize('catl_info', catl_keys_prop_catl_info_arr)
def test_catl_keys_prop_return_type_errors(catl_kind, catl_info, return_type):
    """
    Tests the function:
        cosmo_utils.mock_catalogues.catls_utils.catl_keys_prop` for input and 
    output variables.

    It verifies if errors are raised when `return_type` is incorrect

    Parameters
    -----------
    catl_kind : {'data', 'mocks'} str
        Type of catalogue to use. This variable is set to `data` by default.

        Options:
            - `data` : catalogues come from SDSS `real` catalogue
            - `mocks` : catalogue come from SDSS `mock` catalogues
    """
    ## Running function
    with pytest.raises(LSSUtils_Error):
        output = catls_utils.catl_keys_prop(catl_kind, catl_info=catl_info,
            return_type=return_type)


#########-------------------------------------------------------------#########



















#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-07
# Last Modified: 2018-05-07
from __future__ import print_function, division, absolute_import
__author__     =['Victor Calderon']
__copyright__  =["Copyright 2018 Victor Calderon"]
__email__      =['victor.calderon@vanderbilt.edu']
__maintainer__ =['Victor Calderon']
"""
Set of test functions for the `mags_calculations` functions
"""

## Import modules
import numpy as np
import pytest
from   cosmo_utils.mock_catalogues import mags_calculations
from   cosmo_utils.custom_exceptions import LSSUtils_Error

## Functions

## Testing `apparent_to_absolute_magnitude` function
app_abs_test_arr = [    (-0.72,  -3.1128 , 30.1  , 'pc'),
                        ( 0.14,  -7.0653 , 0.2761, 'kpc'),
                        ( 1.26,  -7.1945 , 0.4908, 'kpc'),
                        ( 6.00,  11.0    , 1     , 'pc'),
                        (-4.0 ,  -4.3959 , 12    , 'pc')]
@pytest.mark.parametrize('app_mag, abs_mag, lum_dist, unit', app_abs_test_arr)
def test_apparent_to_absolute_magnitude(app_mag, abs_mag, lum_dist, unit):
    """
    Tests the function 
    `cosmo_utils.mock_catalogues.mags_calculations.apparent_to_absolute_magnitude`
    for input and output parameters.

    This function tests whether or not the function recovers the 
    expected absolute magnitude.

    Parameters
    ----------
    app_mag : array_like
        Array of apparent magnitude(s)

    abs_mag : np.ndarray
        Array of absolute magnitudes. `abs_mag` is a float if 
        `app_mag` is a float or int.

    lum_dist : array_like
        Array of luminosity distnace to object. In units of `Mpc`.

    unit : {'pc', 'kpc', 'mpc'} str, optional
        Unit to use for `lum_dist`. This variable is set to `mpc` by
        default. When `pc`, the units are in parsecs, while `mpc` is for 
        distances in mega-parsecs, etc.
    """
    ## Calculations
    out_abs_mag = mags_calculations.apparent_to_absolute_magnitude(
        app_mag, lum_dist, unit=unit)
    ## Comparing to expected absolute distance
    np.testing.assert_almost_equal(abs_mag, out_abs_mag, decimal=1)

## Getting Sun's magnitudes
get_sun_mag_test_arr = [    ('U', 'Binney_and_Merrifield_1998', 5.61),
                            ('B', 'Binney_and_Merrifield_1998', 5.48),
                            ('V', 'Binney_and_Merrifield_1998', 4.83),
                            ('R', 'Binney_and_Merrifield_1998', 4.42),
                            ('I', 'Binney_and_Merrifield_1998', 4.08),
                            ('J', 'Binney_and_Merrifield_1998', 3.64),
                            ('H', 'Binney_and_Merrifield_1998', 3.32),
                            ('K', 'Binney_and_Merrifield_1998', 3.28),
                            ('u', 'SDSS_Blanton_2003_z0.1', 6.80),
                            ('g', 'SDSS_Blanton_2003_z0.1', 5.45),
                            ('r', 'SDSS_Blanton_2003_z0.1', 4.76),
                            ('i', 'SDSS_Blanton_2003_z0.1', 4.58),
                            ('z', 'SDSS_Blanton_2003_z0.1', 4.51)]
@pytest.mark.parametrize('filter_opt, system, expected', get_sun_mag_test_arr)
def test_get_sun_abs_mag(filter_opt, system, expected):
    """
    Tests the function 
    `cosmo_utils.mock_catalogues.mags_calculations.apparent_to_absolute_magnitude`
    for input and output parameters.

    This function tests whether or not the function recovers the 
    expected absolute magnitude.

    Parameters
    ----------
    filter: string
        magnitude filter to use.
    
    system: string
        Kind of filter to use. (default = `SDSS_Blanton_2003_z0.1`)
        Options: 
        - `Binney_and_Merrifield_1998`: See Binney and Merrifield 1998
        - `SDSS_Blanton_2003_z0.1`: See Blanton et al. 2003 equation 14

    expected : float
        Expected solar absolute magnitude in `filter` using `system`.
    """
    # Output from function
    out_sun_abs_mag = mags_calculations.get_sun_mag(filter_opt, system=system)
    # Comparing to expected absolute magnitude
    np.testing.assert_almost_equal(expected, out_sun_abs_mag)




# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions.

from __future__ import absolute_import

from cosmo_utils.ml import ml_utils



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-04-28
# Last Modified: 2018-04-28
from __future__ import absolute_import, division, print_function
__author__     = ['Victor Calderon']
__copyright__  = ["Copyright 2018 Victor Calderon"]
__email__      = ['victor.calderon@vanderbilt.edu']
__maintainer__ = ['Victor Calderon']
__all__        = [  "myceil",
                    "myfloor",
                    "Bins_array_create",
                    "sigma_calcs",
                    "Stats_one_arr"]
"""
Set of statistical functions
"""

## Import modules
import math
import numpy as np
from   cosmo_utils.utils             import file_utils as fd
from   cosmo_utils.custom_exceptions import LSSUtils_Error

## Functions

# Upper-bound values
def myceil(x, base=10):
    """
    Determines the upper-bound interger for a given number with a given base.

    Parameters
    ----------
    x : float
        Number to be approximated to closest number to `base`

    base : float
        Base used to calculate the closest largest number

    Returns
    ----------
    y : float
        Closest float number to `x`, i.e. upper-bound float

    Examples
    ----------
    >>> myceil(12, 10)
    20.0

    >>> myceil(12.05, 1.)
    13.0

    >>> myceil(12.05, 0.5)
    12.5
    """
    y = float(base * math.ceil(float(x)/base))

    return y

## Lower-bound values
def myfloor(x, base=10):
    """
    Determines the lower-bound interger for a given number with a given base.

    Parameters
    ----------
    x : float
        Number to be approximated to closest number to `base`

    base : float
        Base used to calculate the closest largest number

    Returns
    ----------
    y : float
        Closest float number to `x`, i.e. upper-bound float

    Examples
    ----------
    >>> myfloor(12, 10)
    10.0

    >>> myfloor(12.05, 1.)
    12.0

    >>> myfloor(12.05, 0.2)
    12.0
    """
    y = float(base * math.floor(float(x)/base))

    return y

## Generation of bins evenly spaced out
def Bins_array_create(arr, base=10):
    """
    Generates an evenly-spaced array between the minimum and maximum value
    of a given array,

    Parameters
    ----------
    arr : array_like
        Array of of numbers or floats

    base : int or float, optional
        Interval used to create the evenly-spaced array of elements

    Returns
    ----------
    bins_arr : `numpy.ndarray`
        Array of elements separated in intervals of `base`
    """
    file_msg = fd.Program_Msg(__file__)
    # Transforming input data
    base = float(base)
    arr = np.asarray(arr)
    # Checking array dimensions
    if arr.ndim != 1:
        msg = '{0} The input array is not of dimension 1, but of `{1}`'.format(
            file_msg, arr.ndim)
        raise LSSUtils_Error(msg)
    # Creating evenly-spaced array
    arr_min  = myfloor(arr.min(), base=base)
    arr_max  = myceil(arr.max(), base=base)
    bins_arr = np.arange(arr_min, arr_max + 0.5*base, base)

    return bins_arr

## Calculations of percentiles and sigmas
def sigma_calcs(data_arr, type_sigma='std', perc_arr=[68., 95., 99.7],
    return_mean_std=False):
    """
    Calcualates the 1-, 2-, and 3-sigma ranges for `data_arr`

    Parameters
    -----------
    data_arr : `numpy.ndarray`, shape( param_dict['nrpbins'], param_dict['itern_tot'])
        array of values, from which to calculate percentiles or St. Dev.

    type_sigma : {'perc', 'std'} string, optional (default = 'std')
        Option for calculating either `percentiles` or `standard deviations`
            - 'perc': calculates percentiles
            - 'std' : uses standard deviations as 1-, 2-, and 3-sigmas

    perc_arr : array_like, optional (default = [68., 95., 99.7])
        Array of percentiles to calculate

    return_mean_std : `bool`, optional (default = False)
        Option for returning mean and St. Dev. along with `sigma_dict`

    Return
    ----------
    sigma_dict: python dicitionary
        dictionary containg the 1-, 2-, and 3-sigma upper and lower
        ranges for `data-arr`

    mark_mean: array_like
        array of the mean value of `data_arr`.
        Only returned if `return_mean_std == True`

    mark_std: array_like
        array of the St. Dev. value of `data_arr`.
        Only returned if `return_mean_std == True`
    """
    file_msg = fd.Program_Msg(__file__)
    ## Checking input variables
    # `data_arr`
    data_arr_valid_types = (np.ndarray, list)
    if not (isinstance(data_arr, data_arr_valid_types)):
        msg = '{0} `data_arr` ({1}) is not a valid type!'.format(
            file_msg, type(data_arr))
        raise LSSUtils_Error(msg)
    else:
        data_arr = np.asarray(data_arr)
    # `type_sigma`
    type_sigma_valid = ['perc', 'std']
    if not (isinstance(type_sigma, str)):
        msg = '{0} `type_sigma` ({1}) is not a valid type!'.format(
            file_msg, type(type_sigma))
        raise LSSUtils_Error(msg)
    if not (type_sigma in type_sigma_valid):
        msg = '{0} `type_sigma` ({1}) is not a valid input choice!'.format(
            file_msg, type_sigma)
    ## Determining shape of `data_arr`
    if data_arr.ndim == 1:
        axis = 0
    else:
        axis = 1
    ## Creating dictionary for saving `sigma`s
    sigma_dict = {}
    for ii in range(len(perc_arr)):
        sigma_dict[ii] = []
    ## Using Percentiles to estimate errors
    if type_sigma == 'perc':
        for ii, perc_ii in enumerate(perc_arr):
            mark_lower = np.nanpercentile(data_arr, 50.-(perc_ii/2.),axis=axis)
            mark_upper = np.nanpercentile(data_arr, 50.+(perc_ii/2.),axis=axis)
            # Saving to dictionary
            sigma_dict[ii] = np.column_stack((mark_lower, mark_upper)).T
    ## Using standard deviations to estimate errors
    if type_sigma == 'std':
        mean_val = np.nanmean(data_arr, axis=axis)
        std_val  = np.nanstd( data_arr, axis=axis)
        for ii in range(len(perc_arr)):
            mark_lower = mean_val - ((ii+1) * std_val)
            mark_upper = mean_val + ((ii+1) * std_val)
            # Saving to dictionary
            sigma_dict[ii] = np.column_stack((mark_lower, mark_upper)).T
    ##
    ## Estimating mean and St. Dev. of `data_arr`
    mark_mean = np.nanmean(data_arr, axis=axis)
    mark_std  = np.nanstd( data_arr, axis=axis)
    ## Fixing values for when `axis == 0`
    if data_arr.ndim == 1:
        for ii in range(len(sigma_dict.keys())):
            sigma_dict[ii] = sigma_dict[ii].flatten()

    if return_mean_std:
        return sigma_dict, mark_mean, mark_std
    else:
        return sigma_dict

## Main framework for `Stats_one_arr` and `Stats_two_arr`
def Stats_one_arr(x, y, base=1., arr_len=0, arr_digit='n',
    weights=None, statfunc=np.nanmean, bin_statval='average',
    return_perc=False, failval=np.nan):
    """
    Calculates statists for 2 arrays

    Parameters
    ----------
    x, y : array_like, shape(N,)
        Sets of elements for the 1st and 2nd observable

    base : float, optional
        Bin width in units of `x`. This variable is set to 1. by default.

    arr_len : int, optional
        Minimum number of elements in each bin of `x`

    arr_digit : {'n', 'y', 'o'} str, optional
        Option for which elements to return.

        Options:
            - 'n' : Returns `x_stat`, `y_stat`, `y_std`, `y_std_err`
            - 'y' : Returns `x_stat`, `y_stat`, `y_std`, `y_std_err`, `x_bins_data`, `y_bins_data`
            - 'o' : Returns `x_bins_data`, `y_bins_data`

    weights : array_like or NoneType, optional
        Array of weights for values in `y`. This is set to None by default.

    statfunc : {`numpy.nanmean`, `numpy.nanmedian`} statistical func, optional
        Numerical function used to calculate on bins of data.
        By default, this variable is set to `numpy.nanmean`

    bin_statval : {'average', 'left', 'right'} str, optional
        Option for where to put the bin values of `x` and `y`.
        By default, this variable is set to `average`, which means
        that the values are those of the averages of the bins in `x` and
        `y`.

    return_perc : `bool`, optional
        If true, it also returns the `percentiles` of the data.
        Last item in the return list.
        This variable is set to False by default.

    failval : int, float, NoneType, or NaN, optional
        This is the value used when no data is available for the bin.
        This is set to `numpy.nan` by default

    Returns
    ----------
    x_stat, y_stat : array_like
        Binned array of elements from `x`

    y_std : array_like
        Standard deviation of the binned array in `x`

    y_std_err : array_like
        Error in the `statfunc` of `y`

    x_bins_data : array_like, optional
        Elements of `x` in each bin with spacing of `base`.
        Only returned if `arr_digit` == 'y' or 'o'

    y_bins_data : array_like, optional
        Elements of `y` in each bin with spacing of `base`.
        Only returned if `arr_digit` == 'y' or 'o'

    perc_lims : array_like, shape(N,3)
        Percentiles in each bin of `x_stat`.
        Only returned if `arr_digit` == 'y' or 'o'
    """
    file_msg = fd.Program_Msg(__file__)
    ## Verifying input values
    # `arr_digit`
    if not ((arr_digit == 'y') or (arr_digit == 'n') or (arr_digit == 'o')):
        msg = '{0} `arr_digit` ({1}) is not a valid input. Exiting'.format(
            file_msg, arr_digit)
        raise LSSUtils_Error(msg)
    # Array dimensions
    if not ((len(x) > 0) and (len(y) > 0)):
        msg = '{0} The arrays `x` and `y` must have at least one value'
        msg = msg.format(file_msg)
        raise LSSUtils_Error(msg)
    if not ((np.asarray(x).ndim == 1) and (np.asarray(y).ndim == 1)):
        msg = '{0} The arrays `x` and `y` must have dimension of `1`'
        msg = msg.format(file_msg)
        raise LSSUtils_Error(msg)
    # `arr_len`
    if not (arr_len >= 0):
        msg = '{0} `arr_len` ({1}) must be greater or equal than zero!'.format(
            file_msg, arr_len)
        raise LSSUtils_Error(msg)
    # `bin_statval`
    if not (bin_statval in ['average', 'left', 'right']):
        msg = '{0} `bin_statval` ({1}) is not a valid input! Exiting'.format(
            file_msg, bin_statval)
        raise LSSUtils_Error(msg)
    ##
    ## Converting arrays to numpy arrays
    x       = np.asarray(x)
    y       = np.asarray(y)
    arr_len = int(arr_len - 1.) if arr_len != 0 else int(arr_len)
    ##
    ## Statistics calculations
    x_bins   = Bins_array_create(x, base=base)
    x_digits = np.digitize(x, x_bins)
    ##
    ## Determining which bins to use
    ## These are the bins that meet the criteria of `arr_len`
    x_digits_bins = np.array([int(ii) for ii in range(1, len(x_bins))
        if len(x_digits[x_digits == ii]) > arr_len])
    ## Elements in each bin
    # X-values
    x_bins_data = np.array([x[x_digits == ii] for ii in x_digits_bins])
    # Y-values
    y_bins_data = np.array([y[x_digits == ii] for ii in x_digits_bins])
    ##
    ## Selecting data in bins
    # Centered around the average
    if (bin_statval == 'average'):
        x_stat = np.array([statfunc(ii) if len(ii) > arr_len else failval
            for ii in x_bins_data])
    # Left-hand side of the bin
    if (bin_statval == 'left'):
        x_stat = np.array([x_bins[:-1][ii] if
            len(x_bins_data[ii]) > arr_len else failval for ii
            in range(len(x_bins_data))])
    # Right-hand side of the bin
    if (bin_statval == 'right'):
        x_stat = np.array([x_bins[1:][ii] if
            len(x_bins_data[ii]) > arr_len else failval for ii
            in range(len(x_bins_data))])
    ##
    ## Determining the values in `y`
    # `stat_function`
    y_stat = np.array([statfunc(ii) if len(ii) > arr_len else failval
        for ii in y_bins_data])
    # Standard Deviation
    y_std  = np.array([np.nanstd(ii) if len(ii) > arr_len else failval
            for ii in y_bins_data])
    # Error in the mean/median
    y_std_err = np.array([
        np.nanstd(ii)/math.sqrt(len(ii)) if len(ii) > arr_len else failval
        for ii in y_bins_data])
    ##
    ## Correcting error inf `statfunc` == `numpy.nanmedian`
    if statfunc == np.nanmedian:
        y_std_err *= 1.253
    ##
    ## Returning percentiles
    if return_perc:
        perc_arr_lims = sigma_calcs(y_stat)
    ##
    ## Returning values
    if return_perc:
        if arr_digit == 'n':
            return_val = [  x_stat, y_stat, y_std, y_std_err, perc_arr_lims]
        if arr_digit == 'y':
            return_val = [  x_stat, y_stat, y_std, y_std_err,
                            x_bins_data, y_bins_data, perc_arr_lims]
        if arr_digit == 'o':
            return_val = [  x_bins_data, y_bins_data, perc_arr_lims]
    else:
        if arr_digit == 'n':
            return_val = [  x_stat, y_stat, y_std, y_std_err]
        if arr_digit == 'y':
            return_val = [  x_stat, y_stat, y_std, y_std_err,
                            x_bins_data, y_bins_data]
        if arr_digit == 'o':
            return_val = [  x_bins_data, y_bins_data]

    return return_val



def test_deprecation():
    import warnings
    warnings.warn(
        "This is deprecated, but shouldn't raise an exception, unless "
        "enable_deprecations_as_exceptions() called from conftest.py",
        DeprecationWarning)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-02
# Last Modified: 2018-05-02
from __future__ import absolute_import, division, print_function
__author__     = ['Victor Calderon']
__copyright__  = ["Copyright 2018 Victor Calderon"]
__email__      = ['victor.calderon@vanderbilt.edu']
__maintainer__ = ['Victor Calderon']
__all__        = [  "IDL_read_file",
                    "fast_food_reader",
                    "read_pandas_hdf5",
                    "read_hdf5_file_to_pandas_DF",
                    "pandas_file_to_hdf5_file",
                    "pandas_df_to_hdf5_file",
                    "concatenate_pd_df"]
"""
Set of functions to read various types of files
"""

# Import modules
import struct
import os
import sys
import numpy as np
import pandas as pd
import h5py
from   scipy.io.idl import readsav
from   cosmo_utils.utils import file_utils as fd
from   cosmo_utils.custom_exceptions import LSSUtils_Error

## Functions

# Reads in IDL catalogue as Python dictionary
def IDL_read_file(idl_file):
    """
    Reads an IDL file and converts it to a Python dictionary

    Parameters
    ----------
    idl_file : string
        Path to the filename being used

    Returns
    ----------
    idl_dict : python dictionary
        Dictionary with the data from `idl_file`
    """
    # Checking that file exists
    fd.File_Exists(idl_file)
    # Converting to dictionary
    try:
        idl_dict = readsav(idl_file, python_dict=True)
    except:
        msg = '{0} `idl_file` {0} is not an IDL file'.format(
            fd.Program_Msg(__file__), idl_file)
        raise LSSUtils_Error(msg)

    return idl_dict

# Reads in `fastfood`-type files and converts it to an array
def fast_food_reader(key, nitems, filename):
    """
    Reads in `fastfood`-type file and converts it to an array

    Parameters
    ----------
    key : {'int', 'float', 'double', 'long'}, str
        Type of the element to extract.

    nitmes : int
        Number of items to expect and exctract.

    filename : str
        Absolute path to the file, from which to extract the information.

    Returns
    ----------
    items_arr : np.ndarray, shape (`nitems`,)
        Array of elements from `filename` with length `nitems`.
    """
    # Dictionaries with values for `key`
    size_types  = { 'int'       :4  , 'float'       :4   , 'char'    :1 ,\
                    'short_int' :2  , 'long_int'    :4   , 'bool'    :1 ,\
                    'double'    :8  , 'long_double' :8   , 'wchart_t':2 }
    ##
    type_string = { 'char'         :'c', 'signed_char'       :'b',\
                    'unsigned_char':'B', '_Bool'             :'?',\
                    'short'        :'h', 'unsigned_short'    :'H',\
                    'int'          :'i', 'unsigned_int'      :'I',\
                    'long'         :'l', 'unsigned_long'     :'L',\
                    'long_long'    :'q', 'unsigned_long_long':'Q',\
                    'float'        :'f', 'double'            :'d' }
    ## Defining types and sizes
    if key == 'int':
        size_type = size_types ['int']
        type_str  = type_string['int']
    if key == 'float':
        size_type = size_types ['float']
        type_str  = type_string['float']
    if key == 'long':
        size_type = size_types ['long_double']
        type_str  = type_string['long']
    if key == 'double':
        size_type = size_types ['double']
        type_str  = type_string['double']
    ##
    # Top padding (it should contain 4 bytes for `.ff` files)
    # 1st padding = nbyte1
    nbyte1_read = filename.read(1 * size_types['int'])
    nbyte1      = struct.unpack(1 * type_string['int'], nbyte1_read)
    nbyte1_val  = int(nbyte1[0])
    if len(nbyte1) != 1:
        errno = -10
        raise ValueError('Read error: file empty?. \nError: ' + str(errno))
        sys.exit()
    # Extracting data
    nitem1_read = filename.read(size_type * nitems)
    items_arr     = struct.unpack(type_str * nitems, nitem1_read)
    val_len     = len(items_arr)
    if val_len != nitems:
        errno = -20
        raise ValueError('Read Error: {0} items expected. Read {1}'.format(
            nitems, val_len))
        sys.exit()
    # Bottom padding
    nbyte2_read = filename.read(size_types['int'])
    nbyte2      = struct.unpack(1 * type_string['int'], nbyte2_read)
    nbyte2_val  = int(nbyte2[0])
    if len(nbyte2) != 1:
        errno = -30
        raise ValueError('Read Error: File too short?\n Errno: ' + str(errno))
        sys.exit()
    # Checking top and bottom
    if nbyte1_val != nbyte2_val:
        errno = -1
        Err_msg = 'Read warning. Byte numbers do not match \n '
        Err_msg += 'nbyte1 = {0}, nbyte2 = {1}\n'.format(nbyte1, nbyte2)
        raise ValueError(Err_msg + 'Errno: {0}'.format(errno))
        sys.exit()
    # Checking that nbye1_val = nitems*Size_type[key]
    if nbyte1_val != nitems * size_type:
        errno = -2
        Err_msg = 'Read warning. Byte numbers do not match \n '
        Err_msg += 'nbyte1 = {0}, nitems = {1}\n'.format(nbyte1, nitems)
        raise ValueError(Err_msg + 'Errno: {0}'.format(errno))
        sys.exit()
    #
    # Converting values to numpy array
    items_arr = np.asarray(items_arr)

    return items_arr

# Reads in a pandas DataFrame from a HDF5 file
def read_pandas_hdf5(hdf5_file, key=None, ret=False):
    """
    Reads a HDF5 file that contains one or many datasets.
    It converts it into a pandas DataFrame.

    Parameters
    ----------
    hdf5_file : str
        Path to the HDF5 file containing one or more pandas DataFrame(s).

    key : str or NoneType
        If provided, it will extract the `key` value as a pandas DataFrame.
        This value is set to `None` by default.

    ret : `bool`, optional
        If True, it returns the value of the `key`.
        By default, it is set to False.

    Returns
    ----------
    df : `pandas.DataFrame`
        DataFrame from the `hdf5_file` with the data from the `key` directory
    """
    file_msg = fd.Program_Msg(__file__)
    # Checking that file exists
    fd.File_Exists(hdf5_file)
    # Checking number of keys
    hdf5_obj  = pd.HDFStore(hdf5_file)
    hdf5_keys = [ii for ii in hdf5_obj.keys()]
    hdf5_obj.close()
    # Reading in HDF5 file
    if key is None:
        try:
            df = pd.read_hdf(hdf5_file)
            if ret:
                return df, hdf5_keys[0]
            else:
                return df
        except:
            msg  = '{0} Must specify which key to use:\n\t'.format(file_msg)
            msg += 'Possible keys: \n'
            print(msg)
            for key_i, name in enumerate(hdf5_keys):
                print('\t Key {0}:  {1}'.format(key_i, name))
    else:
        if key not in hdf5_keys:
            print('{0} Key not in the file: '.format(file_msg))
            print('Possible Keys:\n')
            for key_i, name in enumerate(hdf5_keys):
                print('\t Key {0}:  {1}'.format(key_i, name))
        else:
            df = pd.read_hdf(hdf5_file, key=key)
            if ret:
                return df, key
            else:
                return df

# Reads HDF5 files and converts them to pandas dataframe
def read_hdf5_file_to_pandas_DF(hdf5_file, key=None):
    """
    Reads content of HDF5 file and converts it to a Pandas DataFrame

    Parameters
    ----------
    hdf5_file : str
        Path to the HDF5 file. This is the file that will be converted
        to a pandas DataFrame.

    key : str or NoneType, optional
        Key or path in `hdf5_file` for the pandas DataFrame and the normal
        HDF5 file.

    Returns
    ----------
    df : `pandas.DataFrame`
        DataFrame from `hdf5_file` under the `key` directory.
    """
    file_msg = fd.Program_Msg(__file__)
    fd.File_Exists(hdf5_file)
    # Reading in Pandas DataFrame
    try:
        df = pd.read_hdf(hdf5_file, key=key)
    except:
        msg = '{0} Could not read `hdf5_file` ({1})! Please check if it exists'
        msg = msg.format(file_msg, hdf5_file)
        raise LSSUtils_Error(file_msg)

    return df

# Converts pandas DataFrame to HDF5 file format
def pandas_file_to_hdf5_file(df_file, hdf5_file, key=None, mode='w'):
    """
    Converts a HDF5 with pandas format and converts it to normal HDF5 file

    Paramters
    ---------
    df_file : str
        Path to the `df_file` containing the pandas DataFrame to be converted

    hdf5_file : str
        Path to the output HDF5 file containg arrays as keys

    key : str or NoneType, optional
        Key or path in HDF5 file for the `df_file` and `hdf5_file`
    """
    file_msg = fd.Program_Msg(__file__)
    fd.File_Exists(df_file)
    # Reading in DataFrame
    if not key:
        data, key = read_pandas_hdf5(df_file, key=None, ret=True)
    else:
        data = read_pandas_hdf5(df_file, key=key)
    # Rearranging data
    arr_names   = data.dtypes.index.values
    dtypes_arr  = data.dtypes.values
    dtypes_arr  = np.array([x.str for x in dtypes_arr])
    data_dtypes = np.dtype(zip(arr_names, dtypes_arr))
    dataset     = np.recarray((len(data),), dtype=data_dtypes)
    for name in dataset.dtype.names:
        dataset[name] = data[name]
    # Saving file to HDF5 format
    hdf5_obj = h5py.File(hdf5_file, mode=mode)
    hdf5_obj.create_dataset(key, data=dataset)
    hdf5_obj.close()
    msg = '{0} HDF5 file created: {1}'.format(file_msg, hdf5_file)
    print(msg)

# Saves a pandas DataFrame into a normal or a `pandas` HDF5 file
def pandas_df_to_hdf5_file(df, hdf5_file, key=None, mode='w', complevel=8):
    """
    Saves a `pandas.DataFrame` into a `pandas` HDF5 FILE.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame to be converted and saved into a HDF5 file.

    hdf5_file : str
        Path to the output HDF5 file

    key : str or NoneType, optional
        Key or path, under which `df` will be saved in the `hdf5_file`.

    mode : {'w','a'}, optional
        Mode to handle `hdf5_file`. This value is set to `w` by default,
        which stand for `write`.

    complevel : int, optional
        Level of compression for `hdf5_file`.
        The range of `complevel` is rane(0-9).
        This is set to a default of 8.
    """
    file_msg = fd.Program_Msg(__file__)
    # Saving DataFrame to `hdf5_file`
    try:
        df.to_hdf(hdf5_file, key, mode=mode, complevel=complevel)
        msg = '{0} HDF5 file created: {1}'.format(file_msg, hdf5_file)
        print(msg)
    except:
        msg = '{0} Could not create HDF5 file'.format(file_msg)
        raise LSSUtils_Error(msg)

# Concatenating pandas DataFrame into a single DataFrame
def concatenate_pd_df(directory, filetype='hdf5', foutput=None, outonly=True):
    """
    Concatenates pandas DataFrames into a single DataFrame

    Parameters
    ----------
    directory : str
        Path to the folder containing multiple pandas-HDF5 files

    filetype : str, optional
        File format of the file in `directory` to be read
        This is set to `hdf5` by default.

    foutput : str or NoneType
        If not `None`, it is the basename of the output file in HDF5 format

    outonly : `bool`, optional
        If True, it returns the pandas DataFrame.
        If False, it only saved the concatenated `pandas.DataFrame`.

    Returns
    ----------
    df_conc : `pandas.DataFrame`
        DataFrame containing the combined datasets from the files in
        `directory`.

    Raises
    ----------
    LSSUtils_Error : Exception
        If no files are found in `directory`, it raises an error
        warning about this.
    """
    file_msg = fd.Program_Msg(__file__)
    # Checking that `directory` exists
    if not os.path.exists(directory):
        msg = '{0} `directory` {1} is not a valid path! Exiting!'.format(
            file_msg, directory)
        raise LSSUtils_Error(msg)
    # Concatenating files
    files_arr = fd.index(directory, '.' + filetype, sort=True)
    print('{0} Found `{1}` files'.format(file_msg, files_arr.size))
    if len(files_arr) > 0:
        # Initializing array that contains info
        df_arr = [[] for x in range(len(files_arr))]
        # Looping over HDF5 (pandas) files
        for ii, file_ii in enumerate(files_arr):
            df_arr[ii] = read_pandas_hdf5(file_ii)
        # Concatenating arrays
        df_conc = pd.concat(df_arr, ignore_index=True)
        # Deciding name of resulting output file
        if (foutput is not None) and (type(foutput) == str):
            foutput_file = os.path.join(directory,
                                        '{0}.{1}'.format(foutput, filetype))
            # Saving resulting DataFrame
            pandas_df_to_hdf5_file(df_conc, foutput_file, key='/Main')
            # Checking file exists
            fd.File_Exists(foutput_file)
            print('{0} Output file saved in: {2}'.format(
                file_msg, foutput_file))
        # If only outputting concatenated DataFrame
        if outonly:
            return df_conc
    else:
        msg = '{0} No files in `{1}` with extension `{2}`'.format(file_msg,
                directory, filetype)
        raise LSSUtils_Error(msg)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-17
# Last Modified: 2018-05-17
from __future__ import absolute_import, division, print_function
__author__     = ['Victor Calderon']
__copyright__  = ["Copyright 2018 Victor Calderon"]
__email__      = ['victor.calderon@vanderbilt.edu']
__maintainer__ = ['Victor Calderon']
"""
Set of test functions for the `ml_utils` functions
"""

## Import modules
import numpy as np
import pytest
from   cosmo_utils.ml import ml_utils as cml
from   cosmo_utils.custom_exceptions import LSSUtils_Error

# Functions

## Testing function `data_preprocessing` - Data type
feat_arr_shape_arr = [(10, 3), (20, 1), (10, 5)]
@pytest.mark.parametrize('n_samples, n_features', feat_arr_shape_arr)
def test_data_preprocessing(n_samples, n_features, pre_opt='no'):
    """
    Tests the function `~cosmo_utils.ml.ml_utils.data_preprocessing`
    for input and output parameters

    Parameters
    ------------
    n_samples : int
        Total number of samples. This variable is used to construct the
        array of features with shape [n_samples, n_features]

    n_features : int
        Total number of features. This variable is used to construct the
        array of features with shape [n_samples, n_features]

    pre_opt : {'min_max', 'standard', 'normalize', 'no'} `str`, optional
        Type of preprocessing to do on `feat_arr`.

        Options:
            - 'min_max' : Turns `feat_arr` to values between (0,1)
            - 'standard' : Uses `~sklearn.preprocessing.StandardScaler` method
            - 'normalize' : Uses the `~sklearn.preprocessing.Normalizer` method
            - 'no' : No preprocessing on `feat_arr`
    """
    ## Constructing `feat_arr`
    feat_arr = np.random.random_sample((n_samples, n_features))
    # Checking output
    feat_arr_out = cml.data_preprocessing(feat_arr, pre_opt=pre_opt)
    # Comparing arrays
    np.testing.assert_array_equal(feat_arr_out, feat_arr)

## Testing function `data_preprocessing` - Data type - `feat_arr`
feat_arr_shape_arr = ['test', None, 1, 'abc']
pre_opt_opt_arr    = ['min_max', 'standard', 'normalize', 'no']
@pytest.mark.parametrize('feat_arr', feat_arr_shape_arr)
@pytest.mark.parametrize('pre_opt' , pre_opt_opt_arr)
def test_data_processing_feat_arr_type(feat_arr, pre_opt):
    """
    Tests the function `~cosmo_utils.ml.ml_utils.data_preprocessing`
    for input and output parameters.
    This function tests for the input type of `feat_arr`

    Parameters
    -----------
    feat_arr : `numpy.ndarray`
        Array of feature values. This array is used for training a
        ML algorithm.

    pre_opt : {'min_max', 'standard', 'normalize', 'no'} `str`, optional
        Type of preprocessing to do on `feat_arr`.

        Options:
            - 'min_max' : Turns `feat_arr` to values between (0,1)
            - 'standard' : Uses `~sklearn.preprocessing.StandardScaler` method
            - 'normalize' : Uses the `~sklearn.preprocessing.Normalizer` method
            - 'no' : No preprocessing on `feat_arr`

    Raises
    ----------
    LSSUtils_Error : Exception
        This error gets raised when `ang` is not a digit or a number.
    """
    # Raising error
    with pytest.raises(LSSUtils_Error):
        cml.data_preprocessing(feat_arr, pre_opt)

## Testing function `data_preprocessing` - Data type - `pre_opt`
feat_arr_shape_arr = [np.array((10, 5))]
pre_opt_opt_arr    = ['min_max_2', 1, None, 'testing']
@pytest.mark.parametrize('feat_arr', feat_arr_shape_arr)
@pytest.mark.parametrize('pre_opt' , pre_opt_opt_arr)
def test_data_processing_pre_opt_type(feat_arr, pre_opt):
    """
    Tests the function `~cosmo_utils.ml.ml_utils.data_preprocessing`
    for input and output parameters.
    This function tests for the input type of `pre_opt`

    Parameters
    -----------
    feat_arr : `numpy.ndarray`
        Array of feature values. This array is used for training a
        ML algorithm.

    pre_opt : {'min_max', 'standard', 'normalize', 'no'} `str`, optional
        Type of preprocessing to do on `feat_arr`.

        Options:
            - 'min_max' : Turns `feat_arr` to values between (0,1)
            - 'standard' : Uses `~sklearn.preprocessing.StandardScaler` method
            - 'normalize' : Uses the `~sklearn.preprocessing.Normalizer` method
            - 'no' : No preprocessing on `feat_arr`

    Raises
    ----------
    LSSUtils_Error : Exception
        This error gets raised when `ang` is not a digit or a number.
    """
    # Raising error
    with pytest.raises(LSSUtils_Error):
        cml.data_preprocessing(feat_arr, pre_opt)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-03
# Last Modified: 2018-05-03
from __future__ import absolute_import, division, print_function
__author__     = ['Victor Calderon']
__copyright__  = ["Copyright 2018 Victor Calderon"]
__email__      = ['victor.calderon@vanderbilt.edu']
__maintainer__ = ['Victor Calderon']
__all__        = [  "flip_angles",
                    "Ang_Distance",
                    "Coord_Transformation"]
"""
Set of geometrical definitions for translations, coordinate tranformations,
etc.
"""

## Import modules
import numpy as np
import pandas as pd
from   cosmo_utils.utils import file_utils as fd
from   cosmo_utils.custom_exceptions import LSSUtils_Error

## Functions

# Restricting angles to be between 0 and 360
def flip_angles(ang, unit='deg'):
    """
    Ensures that an angle is always between 0 and 360 degrees.

    Parameters
    ----------
    ang : float, int, `numpy.ndarray`
        Angle in units of degrees.

    unit : {'deg', 'rad'} str, optional
        Unit of the angle. This variable is set to 'deg' by default.
        If 'rad', the output will be in units of radians.

    Returns
    ----------
    ang_out : float
        Convertions of `ang`, ranging from 0 to 360 degrees.

    Raises
    ----------
    LSSUtils_Error : Exception
        This error gets raised when `ang` is not a digit or a number.

    Examples
    ----------
    >>> flip_angles(-50, unit='deg')
    310.0

    >>> flip_angles(110, unit='deg')
    110.0
    """
    file_msg = fd.Program_Msg(__file__)
    # Checking type of `ang`
    valid_types = (float, int, list, np.ndarray)
    if not (isinstance(ang, valid_types)):
        msg = '{0} `ang` ({1}) is not a number! Exiting!'.format(
            file_msg, ang)
        raise LSSUtils_Error(msg)
    # Checking `unit`
    if not (unit.lower() in ['deg', 'rad']):
        msg = '{0} `unit` ({1}) is not a valid option! Exiting!'.format(
            file_msg, unit)
        raise LSSUtils_Error(msg)
    ##
    ## Converting angle
    if isinstance(ang, float) or isinstance(ang, int):
        ang = float(ang)
        # Converting from radians to degrees, if applicable
        if unit == 'rad':
            ang_converted = np.degrees(ang)
        elif unit == 'deg':
            ang_converted = ang
        # Checking the range of `ang`
        if ang_converted < 0:
            ang_converted += 360
        # Converting back to radians, if applicable
        if unit == 'rad':
            ang_final = np.radians(ang_converted)
        elif unit == 'deg':
            ang_final = float(ang_converted)
    else:
        try:
            ang = np.asarray(ang)
            # Converting to degrees, if applicable
            if unit == 'rad':
                ang_converted = np.degrees(ang)
            elif unit == 'deg':
                ang_converted = ang
            # Checking the range of `ang`
            ang_converted = np.asarray([xx if xx > 0 else xx+360 for xx in
                ang_converted])
            # Converting back to radians, if applicable
            if unit == 'rad':
                ang_final = np.radians(ang_converted)
            elif unit == 'deg':
                ang_final = ang_converted
            # Converting to float
            ang_final = ang_final.astype(float)
        except:
            msg = '{0} `ang` could not be converted!'.format(file_msg)
            raise LSSUtils_Error(msg)

    return ang_final

## Calculates the Angular distance between 2 points
def Ang_Distance(ra1, ra2, dec1, dec2, unit='deg', method='haversine'):
    """
    Calculates angular separation between two sets of points with given
    right ascensions and declinations.

    Taken from: https://en.wikipedia.org/wiki/Haversine_formula

    Parameters
    -----------
    ra1, ra2 : float
        Right Ascension of the 1st and 2nd points. Units in `degrees`
        by default.

    dec1, dec2 : float
        Declination of the 1st and 2nd points. Units in `degrees` by default.

    unit : {'dec','rad'} str, optional
        Unit of `ra1`, `ra2`, `dec1`, and `dec2`.
        This will also determine the final unit that outputs this function.

    method : {'haversine', 'astropy'} str, optional
        Method to use in order to calculate angular separation.
        This variable is to by default to the `haversine` method.
        If `astropy`, it will use the astropy framework to determine the
        angular separation.

    Returns
    -----------
    ang_sep : float
        Angular separation between 1st and 2nd point.
        In units of `degrees`.

    Notes
    -----------
    A = 90. - `dec2`
    B = 90. - `dec1`
    D = `ra1` - `ra2`
    c = Angle between two points
    """
    file_msg = fd.Program_Msg(__file__)
    ## Checking input arguments
    # Valid options
    units_valid   = ['deg', 'rad']
    methods_valid = ['haversine', 'astropy']
    # Units
    if not (unit in units_valid):
        msg = '{0} `unit` ({1}) is not a valid argument'.format(
            file_msg, unit)
        raise LSSUtils_Error(msg)
    # Method
    if not (method in methods_valid):
        msg = '{0} `method` ({1}) is not a valid argument'.format(
            file_msg, method)
        raise LSSUtils_Error(msg)
    ##
    ## Flipping angles
    ra1 = flip_angles(ra1, unit=unit, )
    ra2 = flip_angles(ra2, unit=unit)
    ##
    ## Haversine Method
    if method == 'haversine':
        A = np.radians(90. - dec1)
        B = np.radians(90. - dec2)
        D = np.radians(ra1 - ra2 )
        # Distance
        ang_sep  = (np.sin((A - B) * .5))**2.
        ang_sep += np.sin(A) * np.sin(B) * (np.sin(D * .5))**2.
        ang_sep  = np.degrees(2 * np.arcsin(ang_sep**0.5))
    ##
    ## Astropy Method
    if method == 'astropy':
        # Imports
        from astropy.coordinates import SkyCoord
        from astropy import units as u
        # Converting to `units`
        if unit == 'deg':
            unit_opt = u.degree
        elif unit == 'rad':
            unit_opt = u.radians
        # Calculations
        p1 = SkyCoord(ra=ra1, dec=dec1, unit=(unit_opt, unit_opt))
        p2 = SkyCoord(ra=ra2, dec=dec2, unit=(unit_opt, unit_opt))
        # Angular Separation
        ang_sep = p1.separation(p2)
        # Converting to final units
        if unit == 'deg':
            ang_sep = ang_sep.degree
        elif unit == 'rad':
            ang_sep = ang_sep.radians

    return ang_sep

## Coordinate Transformation function
def Coord_Transformation(ra, dec, dist, ra_cen, dec_cen, dist_cen,
    trans_opt=4, return_dict=False, unit='deg'):
    """
    Transforms spherical coordinates (ra, dec, dist) into cartesian
    coordinates.

    Parameters
    -----------
    ra, dec, dist : array_like, shape (N,)
        Arrays of Right Ascension, declination, and distance.
        Units are ['degrees', 'degrees', 'distance_units']

    ra_cen, dec_cen, dist_cen : float, int
        Right Ascension, declination, and distance for the center of
        the coordinates. These correspond to where the corodinates
        `ra`, `dec`, and `dist` will be centered.

    trans_opt : {1, 2, 3, 4} int, optional
        Option for cartesian translation/transformation for elements.
        This variable ist set to `4` by default.

        Options:
            - 1 : No translation involved
            - 2 : Translation to the center point.
            - 3 : Translation `and` rotation to the center point.
            - 4 : Translation and 2 rotaitons about the center point

    return_dict : {True, False}, `bool`, optional
        If `True`, this functions returns 2 dictionaries with `spherical`
        and `cartesian` coordinates.
        If `False`, it returns a `pandas.DataFrame` with the columns.
        This variable is set to `False` by default.

    unit : {'dec','rad'} str, optional
        Unit of `ra1`, `ra2`, `dec1`, and `dec2`.
        This will also determine the final unit that outputs this function.
        This variable is set to `deg` by default.

    Returns
    -----------
    coord_dict (coord_pd) : python dictionary
        Dictionary with spherical and cartesian dictionary of elements
        based on `trans_opt` value. This value is returned if
        `return_dict` is set to `True`. If not, a `pandas.DataFrame` is
        return.
    """
    file_msg = fd.Program_Msg(__file__)
    ## Check types of elements
    # Units
    unit_arr = ['deg', 'rad']
    if not (unit in unit_arr):
        '{0} `unit` ({1}) is not a valid input!'.format(
            file_msg, unit)
    # Valid types
    valid_types = (float, int, np.ndarray, list)
    # Right Ascension
    if not (isinstance(ra, valid_types)):
        msg = '{0} `ra` ({1}) is not a valid type!'.format(
            file_msg, type(ra))
        raise LSSUtils_Error(msg)
    # Declination
    if not (isinstance(dec, valid_types)):
        msg = '{0} `dec` ({1}) is not a valid type!'.format(
            file_msg, type(dec))
        raise LSSUtils_Error(msg)
    # Distance
    if not (isinstance(dist, valid_types)):
        msg = '{0} `dist` ({1}) is not a valid type!'.format(
            file_msg, type(dist))
        raise LSSUtils_Error(msg)
    # trans_opt
    if not (trans_opt in list(range(1, 5))):
        msg = '{0} `trans_opt` ({1}) is not within valid range!'.format(
            file_msg, trans_opt)
        raise LSSUtils_Error(msg)
    ##
    ## Centre's RA, DEC, DIST
    # Right Ascension
    if (isinstance(ra_cen, float)):
        ra_cen = flip_angles(ra_cen)
    else:
        msg = '{0} `ra_cen` ({1}) is not a float!'.format(
            file_msg, type(ra_cen))
        raise LSSUtils_Error(msg)
    # Declination
    if (isinstance(dec_cen, float)):
        dec_cen = flip_angles(dec_cen)
    else:
        msg = '{0} `dec_cen` ({1}) is not a float!'.format(
            file_msg, type(dec_cen))
        raise LSSUtils_Error(msg)
    # Distance
    if (isinstance(dist_cen, float)):
        dist_cen = float(dist_cen)
    else:
        msg = '{0} `dist_cen` ({1}) is not a float!'.format(
            file_msg, type(dist_cen))
        raise LSSUtils_Error(msg)
    ##
    ## Check type of elements
    # Right Ascension
    if (isinstance(ra, float) or isinstance(ra, int)):
        ra = np.array([ra])
    else:
        ra = np.array(ra)
    # Declination
    if (isinstance(dec, float) or isinstance(dec, int)):
        dec = np.array([dec])
    else:
        dec = np.array(dec)
    # Distance
    if (isinstance(dist, float) or isinstance(dist, int)):
        dist = np.array([dist])
    else:
        dist = np.array(dist)
    ##
    ## Converting to desired units
    if unit == 'rad':
        # Right Ascension
        ra_rad     = ra
        ra_cen_deg = np.degrees(ra_cen)
        ra_cen_rad = ra_cen
        # Declination
        dec_rad     = dec
        dec_cen_deg = np.degrees(dec_cen)
        dec_cen_rad = dec_cen
    elif unit == 'deg':
        ra_rad     = np.radians(ra)
        ra_cen_deg = ra_cen
        ra_cen_rad = np.radians(ra_cen)
        # Declination
        dec_rad     = np.radians(dec)
        dec_cen_deg = dec_cen
        dec_cen_rad = np.radians(dec_cen)
    ##
    ## Initializing pandas DataFrame
    dict_keys  = ['ra', 'dec', 'dist']
    coord_dict = dict(zip(dict_keys, np.vstack([ra, dec, dist])))
    ##
    ## Spherical to Cartesian transformation
    ## 1st tranformation
    # Centre
    x_cen = dist_cen * np.cos(ra_cen_rad) * np.cos(dec_cen_rad)
    y_cen = dist_cen * np.sin(ra_cen_rad) * np.cos(dec_cen_rad)
    z_cen = dist_cen * np.sin(dec_cen_rad)
    # All galaxies
    x = dist * np.cos(ra_rad) * np.cos(dec_rad)
    y = dist * np.sin(ra_rad) * np.cos(dec_rad)
    z = dist * np.sin(dec_rad)
    ##
    ## Rotations about z- and x-axis by `tau` and `Omega`
    ## Rotating the points, not the axes
    x1 = x - x_cen
    y1 = y - y_cen
    z1 = z - z_cen
    # Angles
    omega = np.radians(90. - ra_cen_deg )
    tau   = np.radians(90. - dec_cen_deg)
    # Rotations about z-axis by `omega`
    x2 = x1 * np.cos(omega) - y1 * np.sin(omega)
    y2 = x1 * np.sin(omega) + y1 * np.cos(omega)
    z2 = z1.copy()
    # Rotations about X-axis by `tau`
    x3 = x2.copy()
    y3 = y2 * np.cos(tau) - z2 * np.sin(tau)
    z3 = z2 * np.sin(tau) + z2 * np.cos(tau)
    ##
    ## Definining which variables to return
    # No Translation
    if trans_opt == 1:
        coord_dict['x'] = x
        coord_dict['y'] = y
        coord_dict['z'] = z
    # Translation
    if trans_opt == 2:
        coord_dict['x'] = x1
        coord_dict['y'] = y1
        coord_dict['z'] = z1
    # Translation + Rotation
    if trans_opt == 3:
        coord_dict['x'] = x2
        coord_dict['y'] = y2
        coord_dict['z'] = z2
    # Translation + 2 Rotation (centered about the centre)
    if trans_opt == 4:
        coord_dict['x'] = x3
        coord_dict['y'] = y3
        coord_dict['z'] = z3
    ##
    ## Checking what object to return, i.e. DataFrame or python dictionary
    if return_dict:
        return coord_dict
    else:
        coord_pd = pd.DataFrame(coord_dict)
        return coord_pd



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-07
# Last Modified: 2018-05-07
from __future__ import absolute_import, division, print_function
__author__     = ['Victor Calderon']
__copyright__  = ["Copyright 2018 Victor Calderon"]
__email__      = ['victor.calderon@vanderbilt.edu']
__maintainer__ = ['Victor Calderon']
__all__        = [  "abundance_matching_f",
                    "reversed_arrays"]

# Importing modules
import numpy as np
from   scipy.interpolate import interp1d
from   cosmo_utils.utils import file_utils as fd
from   cosmo_utils.custom_exceptions import LSSUtils_Error

# Functions

# Reversed arrays
def reversed_arrays(x, y):
    """
    Determines if arrays increase or decrease monotonically.

    Parameters
    -----------
    x : `numpy.ndarray`
        Array containing the 1st set of values

    y : `numpy.ndarray`
        Array containing the 2nd set of values.

    Return
    -----------
    mono_opt : `bool`
        If True, `x` increases monotonically with increasing `y`.
        If False, `x` decreases monotonically with increasing `y`.

    Raises
    ----------
    LSSUtils_Error : Exception
        Program exception if input parameters are accepted
    """
    file_msg = fd.Program_Msg(__file__)
    # Testing input arguments
    # x-array
    valid_types = (list, np.ndarray)
    if not (isinstance(x, valid_types)):
        msg = '{0} `x` is not a valid type!'.format(file_msg, type(x))
        raise LSSUtils_Error(msg)
    # y-array
    valid_types = (list, np.ndarray)
    if not (isinstance(y, valid_types)):
        msg = '{0} `y` is not a valid type!'.format(file_msg, type(y))
        raise LSSUtils_Error(msg)
    # x- and y-array shapes
    x = np.asarray(x)
    y = np.asarray(y)
    # Checking dimensions
    if not (x.shape == y.shape):
        msg = '{0} The shape of `x` ({1}) and `y` ({2}) are not the same'
        msg = msg.format(file_msg, x.shape, y.shape)
        raise LSSUtils_Error(msg)
    #
    # Checking if arrays increase or decrease monotonically
    x_diff = np.diff(x).sum()
    y_diff = np.diff(y).sum()
    # Monotonically increasing or decreasing
    if (x_diff > 0) and (y_diff > 0):
        mono_opt = True
    else:
        mono_opt = False

    return mono_opt

## Abundance Matching function
def abundance_matching_f(dict1, dict2, volume1=1., volume2=1., reverse=True,
    dens1_opt=False):
    """
    Abundance matching based on 2 quantities.
    It assigns values from `dict2` to elements in `dict1`

    Parameters
    -----------
    dict1 : python dictionary or `numpy.ndarray`
        Dictionary or array of 1st property.

        Keys :
            - `var` : 1st variable to be analysed
            - `dens` : Density array corresponding to `var` elements.
                        Only if `dens` == True.

    dict2 : python dictionary
        dictionary or array of the 2nd property.

        Keys :
            - `var` : 2nd variable to be analyzed
            - `dens` : Density array corresponding to `var` elements.
                        Given if `dens` == True.

    volume1 : float
        Corresponding volume to `dict1`.

    reverse : `bool`, optional
        Determines the relation between `var1` and `var2`.

    dens1_opt : `bool`, optional
        If True, `density` must be calculated.

        Options :
            - `True` : Density is already provided as key for `dict1`.
            - `False` : Density must be calculated.

    Returns
    -----------
    var1_ab : `numpy.ndarray`
        Array of elements matching those of `dict1`, after matching with
        `dict2`.
    """
    file_msg = fd.Program_Msg(__file__)
    # Check types of input paramenters
    valid_types = (list, dict, np.ndarray)
    # `dict1`
    if not (isinstance(dict1, valid_types)):
        msg = '{0} `dict1` ({1}) is not a valid type!'.format(file_msg,
            type(dict1))
    # `dict2`
    if not (isinstance(dict2, dict)):
        msg = '{0} `dict2` must be a dictionary. Its type is `{1}`'.format(
            file_msg, type(dict2))
        raise LSSUtils_Error(msg)
    # 2nd property
    var2  = np.asarray(dict2['var '])
    dens2 = np.asarray(dict2['dens'])
    #
    # `dens1_opt`
    if dens1_opt:
        # 1st Property
        var1   = np.asarray(dict1['var' ])
        dens_1 = np.asarray(dict1['dens'])
    else:
        if (isinstance(dict1, dict)):
            var1 = dict1['var']
        elif (isinstance(dict1, (list, np.ndarray))):
            var1 = dict1.copy()
        #
        # Determining relation between `var1` and `var2`
        mono_opt_1 = reversed_arrays(var1, var2)
        # Monotonically increasing
        if mono_opt_1:
            counts_1 = np.array([np.where(var1 > x)[0].size for x in var1])
        else:
            counts_1 = np.array([np.where(var1 < x)[0].size for x in var1])
        #
        # Determining density of 1st property
        dens_1 = counts_1.astype(float) / volume1
    #
    # Interpolation for 2nd property
    var2_interp = interp1d(dens2, var2, bounds_error=True, assume_sorted=False)
    # Assigning values to property 1
    var1_ab = np.asarray([var2_interp(xx) for xx in dens_1])

    return var1_ab



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions.

from __future__ import absolute_import

from cosmo_utils.mock_catalogues import abundance_matching
from cosmo_utils.mock_catalogues import catls_utils
from cosmo_utils.mock_catalogues import mags_calculations
from cosmo_utils.mock_catalogues import shmr_funcs
from cosmo_utils.mock_catalogues import spherematch
from cosmo_utils.mock_catalogues import pair_counters



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Victor Calderon
# Created      : 2018-05-09
# Last Modified: 2018-05-09
from __future__ import absolute_import, division, print_function
__author__     = ['Victor Calderon']
__copyright__  = ["Copyright 2018 Victor Calderon"]
__email__      = ['victor.calderon@vanderbilt.edu']
__maintainer__ = ['Victor Calderon']
__all__        = [  "url_checker"]

"""
Tools used to interact with web-related objects
"""

## Import modules
import requests
from   cosmo_utils.utils             import file_utils as fd
from   cosmo_utils.custom_exceptions import LSSUtils_Error

## Functions

## Checking if URL is valid
def url_checker(url_str):
    """
    Checks if the URL is valid or not.

    Parameters
    -----------
    url_str : `str`
        URL of the website to evaluate.

    Raises
    ----------
    LSSUtils_Error : `Exception`
        Program exception if input parameters are accepted
    """
    file_msg = fd.Program_Msg(__file__)
    ## Checking input parameters
    if not (isinstance(url_str, str)):
        msg = '{0} `url_str` ({1}) is not a STRING!'.format(file_msg,
            type(url_str))
        raise LSSUtils_Error(msg)
    ##
    ## Checking Website
    request_url = requests.get(url_str)
    if (request_url.status_code != 200):
        msg = '{0} `url_str` ({1}) does not exist!'.format(file_msg, url_str)
        raise LSSUtils_Error(msg)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions.

from __future__ import absolute_import

from cosmo_utils.utils import file_utils
from cosmo_utils.utils import work_paths
from cosmo_utils.utils import file_readers
from cosmo_utils.utils import stats_funcs
from cosmo_utils.utils import geometry
from cosmo_utils.utils import web_utils
